#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Django and native Python packages
import warnings

# 3rd party packages
import numpy as np

# Project imports
from groundhog.general.validation import Validator


VERTICALCAPACITY_UNDRAINED_API = {
    'effective_length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'effective_width': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'su_base': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'su_increase': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'su_above_base': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'base_depth': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'skirted': {'type': 'bool', },
    'base_sigma_v_eff': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'roughness': {'type': 'float', 'min_value': 0.0, 'max_value': 1.0},
    'horizontal_load': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'foundation_inclination': {'type': 'float', 'min_value': -90.0, 'max_value': 90.0},
    'ground_surface_inclination': {'type': 'float', 'min_value': -90.0, 'max_value': 90.0},
    'bearing_capacity_factor': {'type': 'float', 'min_value': 3.0, 'max_value': 12.0},
    'factor_f_override': {'type': 'float', 'min_value': 0.0, 'max_value': 2.0}
}

VERTICALCAPACITY_UNDRAINED_API_ERRORRETURN = {
    'vertical_capacity [kN]': np.nan,
    'Su2 [kPa]': np.nan,
    'K_c [-]': np.nan,
    's_c [-]': np.nan,
    'd_c [-]': np.nan,
    'i_c [-]': np.nan,
    'b_c [-]': np.nan,
    'g_c [-]': np.nan,
    'F [-]': np.nan,
}

@Validator(VERTICALCAPACITY_UNDRAINED_API, VERTICALCAPACITY_UNDRAINED_API_ERRORRETURN)
def verticalcapacity_undrained_api(effective_length, effective_width, su_base, su_increase=0.0, su_above_base=np.nan,
                                   base_depth=0.0, skirted=True, base_sigma_v_eff=0.0, roughness=0.67, horizontal_load=0.0, foundation_inclination=0.0,
                                   ground_surface_inclination=0.0, bearing_capacity_factor=5.14,
                                   factor_f_override=np.nan, **kwargs):
    """
    Calculates the vertical capacity for a shallow foundation in clay with constant or linearly increasing undrained shear strength according to API RP 2GEO.

    The correction factor consists of a shape factor, a depth factor and three inclination factors. In the case of linearly increasing shear strenght, an additional factor F is used which can either be calculated from the foundation roughness (default) or specified directly.

    NOTE:  The relevancy of using the depth factor (:math:`d_c`) should be evaluated in each case. If the installation procedure and/or other foundation aspects, such as scour, do not allow for the required mobilization of shear stresses in the soil above foundation base level, it is recommended that (:math:`d_c`) = 0. In particular, it is recommended that (:math:`d_c`) = 0 if the horizontal load leads to mobilization of significant passive earth pressure between seafloor and foundation base level.

    NOTE: (:math:`H^{\prime}`) in the equation for inclination factor refers to the load applied to the effective area component of the base only. This corresponds to the total lateral load applied to the foundation minus any soil resistance acting on the foundation above skirt tip level as outlined in A.7.2.1, and minus any lateral resistance that may be carried by shearing at skirt tip level outside the effective area. This value is an input to the function.

    NOTE: For embedded foundations, the eccentricity is affected by the load inclination and increases with embedment. The modified eccentricity is found by calculating the intersection between the loading direction and the base depth level.

    NOTE: For skirted foundations, a separate assessment is required to check whether the skirt spacing is small enough to rely on the undrained shear strenght at skirt tip level.

    NOTE: In the assessment for non-skirted, base embedded foundation, the overburden pressure needs to be included in the vertical capacity assessment

    :param effective_length: Effective length of the foundation (:math:`L^{\\prime}`) [:math:`m`]  - Suggested range: 0.0<=effective_length
    :param effective_width: Minimum effective lateral dimension (:math:`B^{\\prime}`) [:math:`m`]  - Suggested range: 0.0<=effective_width
    :param su_base: Undrained shear strength at foundation base level (:math:`S_{uo}`) [:math:`kPa`]  - Suggested range: 0.0<=su_base
    :param su_increase: Linear increase in undrained shear strength (:math:`\\kappa`) [:math:`kPa/m`] (optional, default=0.0) - Suggested range: 0.0<=su_increase
    :param su_above_base: Average undrained shear strength above base level (:math:`s_{u,ave}`) [:math:`kPa`] (optional, default=np.nan) - Suggested range: 0.0<=su_above_base
    :param base_depth: Depth to the base of the foundation (:math:`D`) [:math:`m`] (optional, default=0.0) - Suggested range: 0.0<=base_depth
    :param skirted: Determines whether a foundation is skirted or base-embedded without skirts (optional, default=True)
    :param base_sigma_v_eff: Vertical effective stress at base level. Only used for non-skirted base-embedded foundations (:math:`\\sigma_{vo}^{\\prime}`) [:math:`kPa`] (optional, default=0.0)
    :param roughness: Value for roughness (0.0 for fully smooth, 1.0 for fully rough) (:math:`-`) [:math:`-`] (optional, default=0.67) - Suggested range: 0.0<=roughness<=1.0
    :param horizontal_load: Horizontal load acting on effective area of foundation (:math:`H^{\\prime}`) [:math:`kN`] (optional, default=0.0) - Suggested range: 0.0<=horizontal_load
    :param foundation_inclination: Foundation inclination as defined in figure (:math:`\\nu`) [:math:`deg`] (optional, default=0.0) - Suggested range: -90.0<=foundation_inclination<=90.0
    :param ground_surface_inclination: Ground surface inclination as defined in figure (:math:`\\beta`) [:math:`deg`] (optional, default=0.0) - Suggested range: -90.0<=ground_surface_inclination<=90.0
    :param bearing_capacity_factor: Bearing capacity factor (:math:`N_c`) [:math:`-`] (optional, default=5.14) - Suggested range: 3.0<=bearing_capacity_factor<=12.0
    :param factor_f_override: Direct specification of the factor F (:math:`F`) [:math:`-`] (optional, default=np.nan) - Suggested range: 0.0<=factor_f_override<=2.0

    .. math::
        Q_d = (s_u \\cdot N_c \\cdot K_c) \\cdot A^{\\prime} + \\sigma_{vo}^{\\prime} \\quad \\text{ (constant)}

        Q_d = F \\cdot \\left( s_{uo} \\cdot N_c + \\frac{\\kappa \\cdot B^{\\prime}}{4} \\right) \\cdot K_c \\cdot A^{\\prime}  + \\sigma_{vo}^{\\prime} \\quad \\text{ (linearly increasing) }

        \\text{Exclude the }  \\sigma_{vo}^{\\prime} \\text{ term for skirted foundations}

        K_c = 1 + s_c + d_c  - i_c - b_c - g_c

        \\text{Correction factors constant undrained shear strength}

        s_c = 0.18 \\cdot (1 - 2 \\cdot i_c) \\cdot (B^{\\prime} / L^{\\prime})

        d_c = 0.3 \\cdot \\arctan ( D / B^{\\prime})

        i_c = 0.5 - 0.5 \\cdot [1 - H^{\\prime} / (A^{\\prime} \\cdot s_u)]^{0.5}

        b_c = 2 \\cdot \\nu / (\\pi + 2) \\approx 0.4 \\cdot \\nu

        g_c = 2 \\cdot \\beta / (\\pi + 2) \\approx 0.4 \\cdot \\beta

        \\text{Correction factors linearly increasing undrained shear strength}

        F \\approx a + b \\cdot x - ((c + b \\cdot x)^2 + d^2)^{0.5}

        x = \\frac{\\kappa \\cdot B^{\\prime}}{s_{uo}}

        s_c = s_{cv} (1 - 2 \\cdot i_c) (B^{\\prime} / L^{\\prime})

        d_c = 0.3 \\cdot (s_{u,ave} / s_{u,2}) \\cdot \\arctan( D / B^{\\prime})

        s_{u,2} = F \\cdot (N_c \\cdot s_{uo} + \\kappa \\cdot B^{\\prime} / 4) / N_c

        i_c = 0.5 - 0.5 \\cdot [1 -H^{\\prime}/(A^{\\prime} \\cdot s_u)]^{0.5}

        b_c = 2 \\cdot \\nu / (\\pi + 2) \\approx 0.4 \\cdot \\nu

        g_c = 2 \\cdot \\beta / (\\pi + 2) \\approx 0.4 \\cdot \\beta

        \\text{Correction for horizontal load}

        H^{\\prime} = H_{total} - H_{d,outside} / \\gamma_{sliding} - \\Delta H / \\gamma_{sliding}

    :returns:   Vertical capacity (:math:`Q_d`) [:math:`kN`], Combined correction factor (:math:`K_c`) [:math:`-`], Shape factor (:math:`s_c`) [:math:`-`], Depth factor (:math:`d_c`) [:math:`-`], Load inclination factor (:math:`i_c`) [:math:`-`], Foundation inclination factor (:math:`b_c`) [:math:`-`], Ground inclination factor (:math:`g_c`) [:math:`-`], Correction factor for shear strength increase (:math:`F`) [:math:`-`]

    :rtype: Python dictionary with keys ['vertical_capacity [kN]','K_c [-]','s_c [-]','d_c [-]','i_c [-]','b_c [-]','g_c [-]','F [-]']

    .. figure:: images/api_correction_factor_undrained_linear.png
        :figwidth: 500
        :width: 400
        :align: center

        API correction factors for linear shear strength increase

    .. figure:: images/api_inclination_factors.png
        :figwidth: 500
        :width: 400
        :align: center

        API inclination factor definition

    Reference - API RP 2GEO, 2011. API RP 2GEO Geotechnical and Foundation Design Considerations

    """
    if su_increase == 0.0:
        # Correction factors for constant su
        if horizontal_load > su_base * effective_width * effective_length:
            raise ValueError("Horizontal load exceeds horizontal capacity")
        i_c = 0.5 - 0.5 * max(0.0, 1.0 - horizontal_load / (su_base * effective_length * effective_width)) ** 0.5
        s_c = 0.18 * (1.0 - 2.0 * i_c) * (effective_width / effective_length)
        d_c = 0.3 * np.arctan(base_depth / effective_width)
        b_c = 2.0 * np.radians(foundation_inclination) / (np.pi + 2.0)
        g_c = 2.0 * np.radians(ground_surface_inclination) / (np.pi + 2.0)
        F_factor = np.nan
        su2 = np.nan
    else:
        # Correction factors for linearly increasing su
        dimensionless_increase = su_increase * effective_width / su_base
        if not np.math.isnan(factor_f_override):
            F_factor = factor_f_override
        else:
            if 0.0 <= dimensionless_increase <= 25.0:
                F_smooth = 1.372 + 0.07 * dimensionless_increase - \
                           np.sqrt(((-0.128 + 0.07 * dimensionless_increase) ** 2.0) + ((0.342) ** 2.0))
                F_rough = 2.56 + 0.457 * dimensionless_increase - \
                           np.sqrt(((0.713 + 0.457 * dimensionless_increase) ** 2.0) + ((1.38) ** 2.0))
            else:
                warnings.warn("kB/Suo outside interpolation range, value for kB/Suo=25 is used", Warning)
                F_smooth = 1.372 + 0.07 * 10.0 - \
                           np.sqrt(((-0.128 + 0.07 * 10.0) ** 2.0) + ((0.342) ** 2.0))
                F_rough = 2.56 + 0.457 * 10.0 - \
                           np.sqrt(((0.713 + 0.457 * 10.0) ** 2.0) + ((1.38) ** 2.0))

            F_factor = np.interp(roughness, [0.0, 1.0], [F_smooth, F_rough])

        if 0.0 <= dimensionless_increase <= 10.0:
            s_cv = 0.18 - 0.155 * ((dimensionless_increase)**0.5) + 0.021 * dimensionless_increase
        else:
            warnings.warn("kB/Suo outside interpolation range, value for kB/Suo=10 is used", Warning)
            s_cv = 0.18 - 0.155 * ((10.0)**0.5) + 0.021 * 10.0
        i_c = 0.5 - 0.5 * (1.0 - horizontal_load / (su_base * effective_length * effective_width)) ** 0.5
        s_c = s_cv * (1.0 - 2.0 * i_c) * (effective_width / effective_length)
        su2 = F_factor * (bearing_capacity_factor * su_base + 0.25 * su_increase * effective_width) / bearing_capacity_factor
        if np.math.isnan(su_above_base):
            raise ValueError("Undrained shear strength above base (su_above_base) must be specified for linearly increasing undrained shear strength cases")
        d_c = 0.3 * (su_above_base / su2) * np.arctan(base_depth / effective_width)
        b_c = 2.0 * np.radians(foundation_inclination) / (np.pi + 2.0)
        g_c = 2.0 * np.radians(ground_surface_inclination) / (np.pi + 2.0)

    K_c = 1.0 + s_c + d_c - i_c - b_c - g_c

    if not skirted and base_sigma_v_eff == 0.0:
        warnings.warn("Vertical effective stress at base for base embedded foundation is zero. Specify base_sigma_v_eff"
                      " to take a non-zero value into account")

    if su_increase == 0.0:
        vertical_capacity = su_base * bearing_capacity_factor * K_c * effective_width * effective_length
        if not skirted:
            vertical_capacity += base_sigma_v_eff
    else:
        vertical_capacity = F_factor * (su_base * bearing_capacity_factor + \
                                        0.25 * su_increase * effective_width) * \
                            K_c * effective_length * effective_width
        if not skirted:
            vertical_capacity += base_sigma_v_eff

    return {
        'vertical_capacity [kN]': vertical_capacity,
        'Su2 [kPa]': su2,
        'K_c [-]': K_c,
        's_c [-]': s_c,
        'd_c [-]': d_c,
        'i_c [-]': i_c,
        'b_c [-]': b_c,
        'g_c [-]': g_c,
        'F [-]': F_factor,
    }


VERTICALCAPACITY_DRAINED_API = {
    'vertical_effective_stress': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'effective_friction_angle': {'type': 'float', 'min_value': 20.0, 'max_value': 50.0},
    'effective_unit_weight': {'type': 'float', 'min_value': 3.0, 'max_value': 12.0},
    'effective_length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'effective_width': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'base_depth': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'skirted': {'type': 'bool', },
    'load_inclination': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'foundation_inclination': {'type': 'float', 'min_value': -90.0, 'max_value': 90.0},
    'ground_surface_inclination': {'type': 'float', 'min_value': -90.0, 'max_value': 90.0},
}

VERTICALCAPACITY_DRAINED_API_ERRORRETURN = {
    'vertical_capacity [kN]': np.nan,
    'N_q [-]': np.nan,
    'N_gamma [-]': np.nan,
    'K_q [-]': np.nan,
    'K_gamma [-]': np.nan,
    's_q [-]': np.nan,
    's_gamma [-]': np.nan,
    'd_q [-]': np.nan,
    'd_gamma [-]': np.nan,
    'i_q [-]': np.nan,
    'i_gamma [-]': np.nan,
    'b_q [-]': np.nan,
    'b_gamma [-]': np.nan,
    'g_q [-]': np.nan,
    'g_gamma [-]': np.nan
}

@Validator(VERTICALCAPACITY_DRAINED_API, VERTICALCAPACITY_DRAINED_API_ERRORRETURN)
def verticalcapacity_drained_api(vertical_effective_stress, effective_friction_angle, effective_unit_weight,
                                 effective_length, effective_width, base_depth=0.0, skirted=True, load_inclination=0.0,
                                 foundation_inclination=0.0, ground_surface_inclination=0.0,
                                 **kwargs):
    """
    Calculates the vertical capacity for a shallow foundation in sand with effective friction angle characterized from drained triaxial tests. For constructing an envelope, this value needs to be multiplied by the tangent of the inclination to obtain the H-coordinate of the envelope point.

    The correction factor consists of a shape factor, a depth factor and three inclination factors for both :math:`K_q` and :math:`K_gamma`.

    NOTE: :math:`H/Q` corresponds to the tangent of the load inclination. For constructing an envelope, load inclinations should be varied from 0.0 to 90.0

    NOTE:  The relevancy of using the depth factors should be evaluated in each case. If the installation procedure and/or other foundation aspects, such as scour, do not allow for the required mobilization of shear stresses in the soil above foundation base level, it is recommended that depth factors are set to zero. In particular, it is recommended that d = 0 if the horizontal load leads to mobilization of significant passive earth pressure between seafloor and foundation base level.

    NOTE: :math:`H` in the equation for inclination factor refers to the load applied to the effective area component of the base only. This corresponds to the total lateral load applied to the foundation minus any soil resistance acting on the foundation above skirt tip level as outlined in A.7.2.1, and minus any lateral resistance that may be carried by shearing at skirt tip level outside the effective area.

    NOTE: For embedded foundations, the eccentricity is affected by the load inclination and increases with embedment. The modified eccentricity is found by calculating the intersection between the loading direction and the base depth level.

    NOTE: For non-skirted, base embedded foundations, the contribution of overburden pressure needs to be taken into account. In the equations in the standard, (Nq - 1) needs to be replaced by Nq

    :param vertical_effective_stress: Vertical effect stress at depth corresponding to foundation base (:math:`p_o^{\\prime}`) [:math:`kPa`]  - Suggested range: 0.0<=vertical_effective_stress
    :param effective_friction_angle: Effective friction angle (for appropriate stress level at foundatoin base) (:math:`\\phi^{\\prime}`) [:math:`deg`]  - Suggested range: 20.0<=effective_friction_angle<=50.0
    :param effective_unit_weight: Effective unit weight at foundation base (:math:`\\gamma^{\\prime}`) [:math:`kN/m3`]  - Suggested range: 3.0<=effective_unit_weight<=12.0
    :param effective_length: Effective length of the footing (:math:`L^{\\prime}`) [:math:`m`]  - Suggested range: 0.0<=effective_length
    :param effective_width: Minimum effective lateral dimension (:math:`B^{\\prime}`) [:math:`m`]  - Suggested range: 0.0<=effective_width
    :param base_depth: Depth of the foundation base (:math:`D`) [:math:`m`] (optional, default=0.0) - Suggested range: 0.0<=base_depth
    :param skirted: Determines whether a foundation is skirted or base-embedded without skirts (optional, default=True)
    :param load_inclination: Inclination of the load taking into account horizontal load on effective area of base only (:math:`H^{\\prime}`) [:math:`deg`] (optional, default=0.0) - Suggested range: 0.0<=load_inclination
    :param foundation_inclination:  Foundation inclination as defined in figure (:math:`\\nu`) [:math:`deg`] (optional, default=0.0) - Suggested range: -90.0<=foundation_inclination<=90.0
    :param ground_surface_inclination: Ground surface inclination as defined in figure  (:math:`\\beta`) [:math:`deg`] (optional, default=0.0) - Suggested range: -90.0<=ground_surface_inclination<=90.0

    .. math::
        Q_d^{\\prime} = \\left [ p_o^{\\prime} (N_q - 1) K_q + 0.5 \\gamma^{\\prime} B^{\\prime} N_{\\gamma} K_{\\gamma} \\right ] A^{\\prime}

        \\text{Use }  N_q \\text{ instead of } (N_q - 1) \\text{ for non-skirted, base embedded foundations}

        N_q = \\exp \\left [ \\pi \\tan \\phi^{\\prime} \\right ] (\\tan^2 (45^{\\circ} + \\phi^{\\prime}/2))

        N_{\\gamma} = 1.5 \\left ( N_q - 1 \\right) \\tan \\phi^{\\prime}

        K_q = i_q \\cdot s_q \\cdot d_q \\cdot b_q \\cdot g_q

        K_{\\gamma} = i_{\\gamma} \\cdot s_{\\gamma} \\cdot d_{\\gamma} \\cdot b_{\\gamma} \\cdot g_{\\gamma}

        i_q = \\left [ 1 -0.5 (H/Q) \\right]^5

        i_{\\gamma} = \\left[ 1-0.7 (H/Q) \\right]^5

        s_q = 1+i_q ( B^{\\prime} / L^{\\prime} ) \\sin \\phi^{\\prime}

        s_{\\gamma} = 1 - 0.4 i_{\\gamma} ( B^{\\prime} / L^{\\prime} )

        d_q = 1 + 1.2 (D/B^{\\prime}) \\tan \\phi^{\\prime} (1 - \\sin \\phi^{\\prime})^2

        d_{\\gamma} = 1

        b_q = e^{-2 \\nu \\tan \\phi^{\\prime}}

        b_{\\gamma} = e^{-2.7 \\nu \\tan \\phi^{\\prime}}

        g_q = g_{\\gamma} = (1 - 0.5 \\tan \\beta)^5

    :returns:   Vertical capacity of the footing (:math:`Q_d`) [:math:`kN`], Bearing capacity factor for frictional resistance (:math:`N_q`) [:math:`-`], Bearing capacity factor for unit weight (:math:`N_{\\gamma}`) [:math:`-`], Combined correction factor for frictional resistance (:math:`K_q`) [:math:`-`], Combined correction factor for unit weight (:math:`K_{\\gamma}`) [:math:`-`], Shape factor frictional (:math:`s_q`) [:math:`-`], Shape factor unit weight (:math:`s_{\\gamma}`) [:math:`-`], Depth factor frictional (:math:`d_q`) [:math:`-`], Depth factor unit weight (:math:`d_{\\gamma}`) [:math:`-`], Inclination factor frictional (:math:`i_q`) [:math:`-`], Inclination factor unit weight (:math:`i_{\\gamma}`) [:math:`-`], Foundation inclination factor frictional (:math:`b_q`) [:math:`-`], Foundation inclination factor unit weight (:math:`b_{\\gamma}`) [:math:`-`], Soil surface inclination factor frictional (:math:`g_q`) [:math:`-`], Soil surface inclination factor unit weight (:math:`g_gamma`) [:math:`-`]

    :rtype: Python dictionary with keys ['vertical_capacity [kN]','N_q [-]','N_gamma [-]','K_q [-]','K_gamma [-]','s_q [-]','s_gamma [-]','d_q [-]','d_gamma [-]','i_q [-]','i_gamma [-]','b_q [-]','b_gamma [-]','g_q [-]','g_gamma [-]']


    Reference - API RP 2GEO, 2011. API RP 2GEO Geotechnical and Foundation Design Considerations

    """
    if load_inclination != 0.0:
        effective_width = effective_width - 2.0 * (np.tan(np.radians(load_inclination)) *  base_depth)

    i_q = (1.0 - 0.5 * np.tan(np.radians(load_inclination))) ** 5.0
    i_gamma = (1.0 - 0.7 * np.tan(np.radians(load_inclination))) ** 5.0
    s_q = 1.0 + i_q * (effective_width / effective_length) * np.sin(np.radians(effective_friction_angle))
    s_gamma = 1.0 - 0.4 * i_gamma * (effective_width / effective_length)
    d_q = 1.0 + 1.2 * (base_depth / effective_width) * np.tan(np.radians(effective_friction_angle)) * \
                ((1.0 - np.sin(np.radians(effective_friction_angle))) ** 2.0)
    d_gamma = 1.0
    b_q = np.exp(-2.0 * np.radians(foundation_inclination) * np.tan(np.radians(effective_friction_angle)))
    b_gamma = np.exp(-2.7 * np.radians(foundation_inclination) * np.tan(np.radians(effective_friction_angle)))
    g_q = (1.0 - 0.5 * np.tan(np.radians(ground_surface_inclination))) ** 5.0
    g_gamma = g_q

    K_q = i_q * s_q * d_q * b_q * g_q
    K_gamma = i_gamma * s_gamma * d_gamma * b_gamma * g_gamma

    N_q = (np.exp(np.pi * np.tan(np.radians(effective_friction_angle)))) * \
         ((np.tan(np.radians(45.0 + 0.5 * effective_friction_angle))) ** 2.0)
    N_gamma = 1.5 * (N_q - 1.0) * np.tan(np.radians(effective_friction_angle))
    if skirted:
        vertical_capacity = (vertical_effective_stress * (N_q - 1.0) * K_q + \
                             0.5 * effective_unit_weight * effective_width * N_gamma * K_gamma) * \
                            (effective_width * effective_length)
    else:
        vertical_capacity = (vertical_effective_stress * (N_q) * K_q + \
                             0.5 * effective_unit_weight * effective_width * N_gamma * K_gamma) * \
                            (effective_width * effective_length)
    return {
        'vertical_capacity [kN]': vertical_capacity,
        'N_q [-]': N_q,
        'N_gamma [-]': N_gamma,
        'K_q [-]': K_q,
        'K_gamma [-]': K_gamma,
        's_q [-]': s_q,
        's_gamma [-]': s_gamma,
        'd_q [-]': d_q,
        'd_gamma [-]': d_gamma,
        'i_q [-]': i_q,
        'i_gamma [-]': i_gamma,
        'b_q [-]': b_q,
        'b_gamma [-]': b_gamma,
        'g_q [-]': g_q,
        'g_gamma [-]': g_gamma,
    }

SLIDINGCAPACITY_UNDRAINED_API = {
    'su_base': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'foundation_area': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'su_above_base': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'embedded_section_area': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'soil_reaction_coefficient': {'type': 'float', 'min_value': 1.0, 'max_value': 6.0},
}

SLIDINGCAPACITY_UNDRAINED_API_ERRORRETURN = {
    'sliding_capacity [kN]': np.nan,
    'base_resistance [kN]': np.nan,
    'skirt_resistance [kN]': np.nan
}

@Validator(SLIDINGCAPACITY_UNDRAINED_API, SLIDINGCAPACITY_UNDRAINED_API_ERRORRETURN)
def slidingcapacity_undrained_api(su_base, foundation_area, su_above_base=0.0, embedded_section_area=0.0,
                                  soil_reaction_coefficient=4.0, **kwargs):
    """
    Calculates the undrained sliding capacity for a shallow foundation on clay, the contribution of skirt resistance is taken into account.

    :param su_base: Undrained shear strength at foundation base level (:math:`S_{uo}`) [:math:`kPa`]  - Suggested range: 0.0<=su_base
    :param foundation_area: Actual foundation area (not the effective area!) (:math:`A`) [:math:`m2`]  - Suggested range: 0.0<=foundation_area
    :param su_above_base: Average undrained shear strength along the skirt depth (:math:`S_{u,ave}`) [:math:`kPa`] (optional, default=0.0) - Suggested range: 0.0<=su_skirts
    :param embedded_section_area: Embedded vertical cross-sectional area of foundation (:math:`A_h`) [:math:`m2`] (optional, default=0.0) - Suggested range: 0.0<=embedded_section_area
    :param soil_reaction_coefficient: Soil reaction coefficient Kru. A value of 4 is recommended for full contact. If active soil resistance cannot be relied upon, the factor should be reduced to 2 (:math:`K_{ru}`) [:math:`-`] (optional, default=4.0) - Suggested range: 1.0<=soil_reaction_coefficient<=6.0

    .. math::
        H_d = S_{uo} \\cdot A

        \\Delta H = K_{ru} \\cdot (S_{u,ave}) \\cdot A_h

    :returns:   Sliding capacity (combined) (:math:`H_d + \\Delta H`) [:math:`kN`], Sliding resistance on the foundation base (:math:`H_d`) [:math:`kN`], Sliding resistance due to active and passive resistance of the skirts (:math:`\\Delta H`) [:math:`kN`]

    :rtype: Python dictionary with keys ['sliding_capacity [kN]','base_resistance [kN]','skirt_resistance [kN]']


    Reference - API RP 2GEO, 2011. API RP 2GEO Geotechnical and Foundation Design Considerations

    """
    base_resistance = su_base * foundation_area
    skirt_resistance = soil_reaction_coefficient * su_above_base * embedded_section_area
    sliding_capacity = base_resistance + skirt_resistance

    return {
        'sliding_capacity [kN]': sliding_capacity,
        'base_resistance [kN]': base_resistance,
        'skirt_resistance [kN]': skirt_resistance,
    }


SLIDINGCAPACITY_DRAINED_API = {
    'vertical_load': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'effective_friction_angle': {'type': 'float', 'min_value': 20.0, 'max_value': 50.0},
    'effective_unit_weight': {'type': 'float', 'min_value': 3.0, 'max_value': 12.0},
    'embedded_section_area': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'depth_to_base': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'reaction_factor_override': {'type': 'float', 'min_value': 0.0, 'max_value': None},
}

SLIDINGCAPACITY_DRAINED_API_ERRORRETURN = {
    'sliding_capacity [kN]': np.nan,
    'base_capacity [kN]': np.nan,
    'skirt_capacity [kN]': np.nan,
    'K_rd [-]': np.nan,
    'K_p [-]': np.nan
}

@Validator(SLIDINGCAPACITY_DRAINED_API, SLIDINGCAPACITY_DRAINED_API_ERRORRETURN)
def slidingcapacity_drained_api(vertical_load, effective_friction_angle, effective_unit_weight,
                                embedded_section_area=0.0, depth_to_base=0.0, reaction_factor_override=np.nan,
                                **kwargs):
    """
    Calculates the drained sliding capacity for a shallow foundation. The base resistance is increased for passive and active soil resistance derived from skirts.

    :param vertical_load: Actual vertical load during relevant loading condition (:math:`Q`) [:math:`kN`]  - Suggested range: 0.0<=vertical_load
    :param effective_friction_angle: Effective friction angle (for appropriate stress level at foundation base) (:math:`\\phi^{\\prime}`) [:math:`deg`]  - Suggested range: 20.0<=effective_friction_angle<=50.0
    :param effective_unit_weight: Effective unit weight (:math:`\\gamma^{\\prime}`) [:math:`kN/m3`]  - Suggested range: 3.0<=effective_unit_weight<=12.0
    :param embedded_section_area: Embedded vertical cross-sectional area of foundation (:math:`A_h`) [:math:`m2`] (optional, default=0.0) - Suggested range: 0.0<=embedded_section_area
    :param depth_to_base: Depth below seafloor to base level (:math:`D_b`) [:math:`m`] (optional, default=0.0) - Suggested range: 0.0<=depth_to_base
    :param reaction_factor_override: Drained horizontal reaction factor (:math:`K_{rd}`) [:math:`-`] (optional, default=np.nan) - Suggested range: 0.0<=reaction_factor_override

    .. math::
        H_d^{\\prime} = Q \\cdot \\tan \\phi^{\\prime}

        \\Delta H = 0.5 \\cdot K_{rd} \\cdot \\gamma^{\\prime} \\cdot D_b \\cdot A_h

        K_{rd} = K_p - (1/K_p)

        K_p = \\tan^2 (45^{\\circ} + 0.5 \\phi^{\\prime})

    :returns:   Combined sliding capacity due to base and skirt resistance (:math:`H_d + \\Delta H`) [:math:`kN`], Sliding resistance due to base friction (:math:`H_d`) [:math:`kN`], Sliding resistance due to active and passive resistance of the skirts (:math:`\\Delta H`) [:math:`kN`], Reaction factor (:math:`K_{rd}`) [:math:`-`], Passive resistance factor (:math:`K_p`) [:math:`-`]

    :rtype: Python dictionary with keys ['sliding_capacity [kN]','base_capacity [kN]','skirt_capacity [kN]','K_rd [-]','K_p [-]']


    Reference - API RP 2GEO, 2011. API RP 2GEO Geotechnical and Foundation Design Considerations
    
    """
    base_capacity = vertical_load * np.tan(np.radians(effective_friction_angle))
    K_p = (np.tan(np.radians(45.0 + 0.5 * effective_friction_angle))) ** 2.0
    if np.math.isnan(reaction_factor_override):
        K_rd = K_p - (1.0 / K_p)
    else:
        K_rd = reaction_factor_override
    skirt_capacity = 0.5 * K_rd * effective_unit_weight * depth_to_base * embedded_section_area
    sliding_capacity = base_capacity + skirt_capacity

    return {
        'sliding_capacity [kN]': sliding_capacity,
        'base_capacity [kN]': base_capacity,
        'skirt_capacity [kN]': skirt_capacity,
        'K_rd [-]': K_rd,
        'K_p [-]': K_p,
    }

EFFECTIVEAREA_RECTANGLE_API = {
    'length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'width': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'vertical_load': {'type': 'float', 'min_value': 0.001, 'max_value': None},
    'moment_length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'moment_width': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'eccentricity_length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'eccentricity_width': {'type': 'float', 'min_value': 0.0, 'max_value': None},
}

EFFECTIVEAREA_RECTANGLE_API_ERRORRETURN = {
    'effective_area [m2]': np.nan,
    'effective_length [m]': np.nan,
    'effective_width [m]': np.nan,
    'eccentricity_length [m]': np.nan,
    'eccentricity_width [m]': np.nan
}

@Validator(EFFECTIVEAREA_RECTANGLE_API, EFFECTIVEAREA_RECTANGLE_API_ERRORRETURN)
def effectivearea_rectangle_api(length, width, vertical_load=np.nan, moment_length=np.nan, moment_width=np.nan,
                                eccentricity_length=np.nan, eccentricity_width=np.nan, **kwargs):
    """
    Calculates the reduced area of a rectangular footing to account for eccentricty of the load. Eccentricities can either be specified from a moment or defined directly.

    NOTE: In the assessment of eccentricity for shallow foundations on undrained soil, the V can include the weight of soil plug inside the skirts

    :param length: Longest foundation dimension (:math:`L`) [:math:`m`]  - Suggested range: 0.0<=length
    :param width: Shortest foundation dimension (:math:`B`) [:math:`m`]  - Suggested range: 0.0<=width
    :param vertical_load: Actual vertical load during relevant loading condition (:math:`Q`) [:math:`kN`] (optional, default=np.nan) - Suggested range: 0.001<=vertical_load
    :param moment_length: Overturning moment aligned with longest foundation dimension (:math:`M_1`) [:math:`kNm`] (optional, default=np.nan) - Suggested range: 0.0<=moment_length
    :param moment_width: Overturning moment aligned with shortest foundation dimension (:math:`M_2`) [:math:`kNm`] (optional, default=np.nan) - Suggested range: 0.0<=moment_width
    :param eccentricity_length: Eccentricity (direct specification) in the longest foundation direction (:math:`e_1`) [:math:`m`] (optional, default=np.nan) - Suggested range: 0.0<=eccentricity_length
    :param eccentricity_width: Eccentricity (direct specification) in the shortest foundation direction (:math:`e_2`) [:math:`m`] (optional, default=np.nan) - Suggested range: 0.0<=eccentricity_width

    .. math::
        e_1 = \\frac{M_1}{Q}

        e_2 = \\frac{M_2}{Q}

        L^{\\prime} = L - 2 \\cdot e_1

        B^{\\prime} = B - 2 \\cdot e_2

        A^{\\prime} = B^{\\prime} \\cdot L^{\\prime}


    :returns:   Effective area used to calculate mudmat capacity (:math:`A^{\\prime}`) [:math:`m2`], Effective length (:math:`L^{\\prime}`) [:math:`m`], Effective width (:math:`B^{\\prime}`) [:math:`m`], Eccentricity in the length direction (:math:`e_1`) [:math:`m`], Eccentricity in the width direction (:math:`e_2`) [:math:`m`]

    :rtype: Python dictionary with keys ['effective_area [m2]','effective_length [m]','effective_width [m]','eccentricity_length [m]','eccentricity_width [m]']

    .. figure:: images/api_reduced_area_rectangle.png
        :figwidth: 500
        :width: 400
        :align: center

        Effective area of a rectangular mudmat

    Reference - API RP 2GEO, 2011. API RP 2GEO Geotechnical and Foundation Design Considerations
    
    """
    if (not np.math.isnan(vertical_load)) and (not np.math.isnan(moment_length)) and \
            (not np.math.isnan(moment_width)) and np.math.isnan(eccentricity_length) and \
            np.math.isnan(eccentricity_width):
        e1 = moment_length / vertical_load
        e2 = moment_width / vertical_load
    elif (not np.math.isnan(eccentricity_length)) and (not np.math.isnan(eccentricity_width)) and \
            np.math.isnan(moment_length) and np.math.isnan(moment_width) and \
            np.math.isnan(vertical_load):
        e1 = eccentricity_length
        e2 = eccentricity_width
    else:
        raise ValueError("Eccentricity needs to be defined through moments or direct specification")
    effective_length = length - 2.0 * e1
    effective_width = width - 2.0 * e2
    effective_area = effective_length * effective_width

    return {
        'effective_area [m2]': effective_area,
        'effective_length [m]': effective_length,
        'effective_width [m]': effective_width,
        'eccentricity_length [m]': eccentricity_length,
        'eccentricity_width [m]': eccentricity_width,
    }

EFFECTIVEAREA_CIRCLE_API = {
    'foundation_radius': {'type': 'float', 'min_value': 0.01, 'max_value': None},
    'vertical_load': {'type': 'float', 'min_value': 0.01, 'max_value': None},
    'overturning_moment': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'eccentricity': {'type': 'float', 'min_value': 0.0, 'max_value': None},
}

EFFECTIVEAREA_CIRCLE_API_ERRORRETURN = {
    'effective_area [m2]': np.nan,
    'effective_length [m]': np.nan,
    'effective_width [m]': np.nan,
    's [m2]': np.nan,
    'eccentricity [m]': np.nan
}

@Validator(EFFECTIVEAREA_CIRCLE_API, EFFECTIVEAREA_CIRCLE_API_ERRORRETURN)
def effectivearea_circle_api(foundation_radius, vertical_load=np.nan, overturning_moment=np.nan, eccentricity=np.nan,
                             **kwargs):
    """
    Calculates the reduced area for a circular foundation to account for load eccentricity. Eccentricity can either be specified through an overturning moment or with direct specification of eccentricity.

    NOTE: In the assessment of eccentricity for shallow foundations on undrained soil, the V can include the weight of soil plug inside the skirts

    :param foundation_radius: Radius of the circular foundation (:math:`R`) [:math:`m`]  - Suggested range: 0.01<=foundation_radius
    :param vertical_load: Actual vertical load during relevant loading condition (:math:`Q`) [:math:`kN`] (optional, default=np.nan) - Suggested range: 0.01<=vertical_load
    :param overturning_moment: Overturning moment acting on the foundation (:math:`M`) [:math:`kNm`] (optional, default=np.nan) - Suggested range: 0.0<=overturning_moment
    :param eccentricity: Eccentricity (direct specification) (:math:`e_2`) [:math:`m`] (optional, default=np.nan) - Suggested range: 0.0<=eccentricity

    .. math::
        A^{\\prime} = 2s=B^{\\prime} L^{\\prime}

        L^{\\prime}=(2s \\sqrt{\\frac{R+e_2}{R-e_2}})^{1/2}

        B^{\\prime} = L^{\\prime} \\sqrt{\\frac{R-e_2}{R+e_2}}

        s = \\frac{\\pi R^2}{2}-[e_2 \\cdot (\\sqrt{R^2-e_2^2}) + R^2 \\arcsin(\\frac{e_2}{R})]

    :returns:   Effective area used for capacity calculation (:math:`A^{\\prime}`) [:math:`m2`], Effective length  (:math:`L^{\\prime}`) [:math:`m`], Effective width (:math:`B^{\\prime}`) [:math:`m`], Parameter s (:math:`s`) [:math:`m2`], Eccentricity used for the calculation (:math:`e_2`) [:math:`m`]

    :rtype: Python dictionary with keys ['effective_area [m2]','effectve length [m]','effective_width [m]','s [m2]','eccentricity [m]']

    .. figure:: images/api_reduced_area_circle.png
        :figwidth: 500
        :width: 400
        :align: center

        Reduced area for a circular foundation

    Reference - API RP 2GEO, 2011. API RP 2GEO Geotechnical and Foundation Design Considerations

    """
    if not np.math.isnan(overturning_moment) and (not np.math.isnan(vertical_load)) and \
            np.math.isnan(eccentricity):
        e2 = overturning_moment / vertical_load
    elif np.math.isnan(overturning_moment) and np.math.isnan(vertical_load) and \
            not np.math.isnan(eccentricity):
        e2 = eccentricity
    else:
        raise ValueError("Eccentricity needs to be specified either through moment or direct specification")
    s_ecc = 0.5 * np.pi * (foundation_radius ** 2.0) - \
            (e2 * (np.sqrt((foundation_radius ** 2.0) - (e2 ** 2.0))) + \
             (foundation_radius ** 2.0) * np.arcsin(e2 / foundation_radius))
    effective_length = (2.0 * s_ecc * np.sqrt((foundation_radius + e2) / (foundation_radius - e2))) ** 0.5
    effective_width = effective_length * np.sqrt((foundation_radius - e2) / (foundation_radius + e2))
    effective_area = effective_length * effective_width

    return {
        'effective_area [m2]': effective_area,
        'effective_length [m]': effective_length,
        'effective_width [m]': effective_width,
        's [m2]': s_ecc,
        'eccentricity [m]': eccentricity,
    }


ENVELOPE_DRAINED_API = {
    'vertical_effective_stress': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'effective_friction_angle_bearing': {'type': 'float', 'min_value': 20.0, 'max_value': 50.0},
    'effective_friction_angle_sliding': {'type': 'float', 'min_value': 15.0, 'max_value': 45.0},
    'effective_unit_weight': {'type': 'float', 'min_value': 3.0, 'max_value': 12.0},
    'effective_length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'effective_width': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'width': {'type': 'float', 'min_value': 0.0, 'max_value': None}
}

ENVELOPE_DRAINED_API_ERRORRETURN = {
    'envelope_v [kN]': None,
    'envelope_h [kN]': None,
    'envelope_h_unchanged [kN]': None,
    'sliding_cuttoff [kN]': None
}

@Validator(ENVELOPE_DRAINED_API, ENVELOPE_DRAINED_API_ERRORRETURN)
def envelope_drained_api(vertical_effective_stress,
                         effective_friction_angle_bearing, effective_friction_angle_sliding, effective_unit_weight,
                         effective_length, effective_width,
                         length, width, **kwargs):
    """
    Calculates a drained failure envelope for shallow foundations according to API RP 2GEO. Note that optional keyword arguments can be specified as documented in the function verticalcapacity_drained_api.

    The envelope is calculated be varying the inclination of the load. Note that the equation for inclination factor will become negative for large inclinations. These values are filtered from the envelope.

    We derive the ultimate horizontal load in bearing from the equation for inclination but need to remember that this should only account for the horizontal capacity on the effective area.

    :param vertical_effective_stress: Vertical effect stress at depth corresponding to foundation base (:math:`p_o^{\\prime}`) [:math:`kPa`]  - Suggested range: 0.0<=vertical_effective_stress
    :param effective_friction_angle_bearing: Effective friction angle for bearing failure (for appropriate stress level at foundation base) (:math:`\\phi^{\\prime}`) [:math:`deg`]  - Suggested range: 20.0<=effective_friction_angle<=50.0
    :param effective_friction_angle_sliding: Effective friction angle for sliding failure (for appropriate stress level at foundatoin base) (:math:`\\delta^{\\prime}`) [:math:`deg`]  - Suggested range: 15.0<=effective_friction_angle<=45.0
    :param effective_unit_weight: Effective unit weight at foundation base (:math:`\\gamma^{\\prime}`) [:math:`kN/m3`]  - Suggested range: 3.0<=effective_unit_weight<=12.0
    :param effective_length: Effective length of the footing (:math:`L^{\\prime}`) [:math:`m`]  - Suggested range: 0.0<=effective_length
    :param effective_width: Minimum effective lateral dimension (:math:`B^{\\prime}`) [:math:`m`]  - Suggested range: 0.0<=effective_width
    :param length: True length of the footing (:math:`L`) [:math:`m`]  - Suggested range: 0.0<=length
    :param width: Minimum true lateral dimension (:math:`B`) [:math:`m`]  - Suggested range: 0.0<=width


    .. math::
        \\tan(\\text{inclination}) = \\frac{H_{eff}}{Q} = \\frac{H - H_{text{outside eff} - \\Delta H}{Q}

    :returns: Dictionary with the following keys:

        - 'envelope_v [kN]': Vertical capacities for the envelope (:math:`V`)  [:math:`kN`]
        - 'envelope_h [kN]': Horizontal capacities for the envelope (:math:`H`)  [:math:`kN`]
        - 'envelope_h_unchanged [kN]': Horizontal capacities for the envelope without accounting for effective area component only (:math:`H`)  [:math:`kN`]

    Reference - API RP 2GEO

    """

    #TODO: TAKE INTO ACCOUNT ECCENTRICITY FOR INCLINED LOAD + SKIRTS

    _inclinations = np.linspace(0.0, 90.0, 100)
    _envelope_v = np.array(list(map(lambda inclination: verticalcapacity_drained_api(
        vertical_effective_stress=vertical_effective_stress,
        effective_friction_angle=effective_friction_angle_bearing,
        effective_unit_weight=effective_unit_weight,
        effective_length=effective_length,
        effective_width=effective_width,
        load_inclination=inclination,
        **kwargs)['vertical_capacity [kN]'], _inclinations)))

    _inclinations = _inclinations[np.where(np.array(_envelope_v) > 0.0)]
    _envelope_v = _envelope_v[np.where(_envelope_v > 0.0)]

    _h_max = []
    _h_base = []
    _h_base_eff = []
    _h_base_outside_eff = []
    _h_skirt = []

    for v in _envelope_v:
        s = slidingcapacity_drained_api(
            vertical_load=v,
            effective_friction_angle=effective_friction_angle_sliding,
            effective_unit_weight=effective_unit_weight,
            **kwargs)
        _h_max.append(s['sliding_capacity [kN]'])
        base = s['base_capacity [kN]']
        base_eff = (base * effective_length * effective_width) / (length * width)
        base_outside_eff = base - base_eff
        _h_base.append(base)
        _h_base_eff.append(base_eff)
        _h_base_outside_eff.append(base_outside_eff)
        _h_skirt.append(s['skirt_capacity [kN]'])

    _envelope_h = np.array(_envelope_v) * np.tan(np.radians(_inclinations)) + np.array(_h_base_outside_eff) + \
        np.array(_h_skirt)
    _envelope_h_unchanged = np.array(_envelope_v) * np.tan(np.radians(_inclinations))


    return {
        'envelope_v [kN]': _envelope_v,
        'envelope_h [kN]': _envelope_h,
        'envelope_h_unchanged [kN]': _envelope_h_unchanged,
        'sliding_cuttoff [kN]': _h_max
    }


ENVELOPE_UNDRAINED_API = {
    'su_base': {'type': 'float', 'min_value': 0.0, 'max_value': 1000.0},
    'length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'width': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'effective_length': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'effective_width': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'factor_sliding': {'type': 'float', 'min_value': 1.0, 'max_value': None},
    'factor_bearing': {'type': 'float', 'min_value': 1.0, 'max_value': None},
}

ENVELOPE_UNDRAINED_API_ERRORRETURN = {
    'envelope_v_unfactored [kN]': None,
    'envelope_h_unfactored [kN]': None,
    'envelope_v_factored [kN]': None,
    'envelope_h_factored [kN]': None,
    'envelope_v_uncorrected [kN]': None,
    'envelope_h_uncorrected [kN]': None,
    'sliding_capacity': None,
    'bearing_capacity': None,
}


@Validator(ENVELOPE_UNDRAINED_API, ENVELOPE_UNDRAINED_API_ERRORRETURN)
def envelope_undrained_api(
        su_base, length, width, effective_length, effective_width,
        factor_sliding=1.5, factor_bearing=2.0, **kwargs):
    """
    Calculates the undrained failure envelope according to API RP 2GEO. It is important to note than only the horizontal load acting on the effective area is taken into account for the envelope. To achieve this, the horizontal load points for the envelope are selected between 0 and the total horizontal capacity (including contributions from skirt resistance and total base area). Subsequently, a load inclination is calculated using only the effective area component of this horizontal load (subtract skirt resistance and horizontal capacity outside effective area). This inclination is used in the vertical capacity equation (through the inclination factor).

The envelope is calculated twice, first without correction for additional eccentricity at base level and next with a correction for this additional eccentricity. This two step approach is required since the load inclination is not known a priori.

To override the behaviour of the bearing and sliding capacity functions, use the optional keywords arguments defined in the function definitions of slidingcapacity_undrained_api and verticalcapacity_undrained_api

    :param su_base: Undrained shear strength at the foundation base (:math:`S_{uo}`) [:math:`kPa`] - Suggested range: 0.0 <= su_base <= 1000.0
    :param length: Total foundation length (:math:`L`) [:math:`m`] - Suggested range: length >= 0.0
    :param width: Total foundation with (:math:`B`) [:math:`m`] - Suggested range: width >= 0.0
    :param effective_length: Effective length (:math:`L^{\\prime}`) [:math:`m`] - Suggested range: effective_length >= 0.0
    :param effective_width: Effective width (:math:`B^{\\prime}`) [:math:`m`] - Suggested range: effective_width >= 0.0
    :param factor_sliding: Resistance factor for sliding, applied to the H component of the envelope (:math:`\\gamma_{sliding}`) [:math:`-`] - Suggested range: factor_sliding >= 1.0 (optional, default= 1.5)
    :param factor_bearing: Resistance factor for bearing failure, applied to the V component of the envelope (:math:`\\gamma_{bearing}`) [:math:`-`] - Suggested range: factor_bearing >= 1.0 (optional, default= 2.0)

    .. math::
        \\Delta e = D \\cdot \\tan \\theta

    :returns: Dictionary with the following keys:

        - 'envelope_v_unfactored [kN]': List with unfactored vertical capacities for the envelope (:math:`V`)  [:math:`kN`]
        - 'envelope_h_unfactored [kN]': List with unfactored horizontal capacities for the envelope (:math:`H`)  [:math:`kN`]
        - 'envelope_v_factored [kN]': List with factored vertical capacities for the envelope (:math:`V_{factored}`)  [:math:`kN`]
        - 'envelope_h_factored [kN]': List with factored horizontal capacities for the envelope (:math:`H_{factored}`)  [:math:`kN`]
        - 'envelope_v_uncorrected [kN]': List with vertical capacities for the envelope, not corrected for the additional eccentricity (:math:`V_{uncorrected}`)  [:math:`kN`]
        - 'envelope_h_uncorrected [kN]': List with horizontal capacities for the envelope, not corrected for the additional eccentricity (:math:`H_{uncorrected}`)  [:math:`kN`]
        - 'sliding_capacity': Dictionary with details for the sliding capacity calculation
        - 'bearing_capacity': Dictionary with details for the bearing capacity calculation for purely vertical load

    .. figure:: images/envelope_undrained_api_1.png
        :figwidth: 500.0
        :width: 450.0
        :align: center

        Undrained failure envelope

    Reference - API RP 2GEO, 2011. API RP 2GEO Geotechnical and Foundation Design Considerations

    """

    # Calculate the sliding capacity
    _sliding_capacity = slidingcapacity_undrained_api(
        su_base=su_base, foundation_area=length * width, **kwargs)
    # Select horizontal loads between zero and the total sliding capacity
    _horizontal_load = np.linspace(0.0, _sliding_capacity['sliding_capacity [kN]'], 100)
    # Calculate the sliding capacity outside effective area
    _h_outside = _sliding_capacity['base_resistance [kN]'] * \
        (1.0 - ((effective_length * effective_width) / (length * width)))
    # Subtract factored sliding capacity outside effective area and factored skirt resistance from horizontal load
    _horizontal_load_corrected = np.maximum(
        np.zeros(_horizontal_load.__len__()),
        _horizontal_load - \
        (_h_outside * factor_sliding) - \
        (_sliding_capacity['skirt_resistance [kN]'] * factor_sliding))
    # Calculate the vertical bearing capacity
    _bearing_capacity = verticalcapacity_undrained_api(
        effective_length=effective_length,
        effective_width=effective_width,
        su_base=su_base,
        **kwargs)
    # Calculate the envelope (first iteration)
    _envelope_v_raw = list(map(lambda h: verticalcapacity_undrained_api(
        effective_length=effective_length,
        effective_width=effective_width,
        su_base=su_base,
        horizontal_load=h,
        **kwargs)['vertical_capacity [kN]'], _horizontal_load_corrected))
    _envelope_v_uncorrected = np.append(_envelope_v_raw, [0.0, ])
    _envelope_h_uncorrected = np.append(_horizontal_load, [_horizontal_load.max(), ])
    try:
        # Correct the envelope for the additional eccentricity
        base_depth = kwargs['base_depth']
        if base_depth != 0.0:
            inclination = np.arctan(_envelope_h_uncorrected / _envelope_v_uncorrected)
            _effective_width_corrected = np.maximum(
                np.zeros(inclination.__len__()),
                effective_width - (2.0 * base_depth * np.tan(inclination)))
            envelope_v_corrected_raw = list(map(lambda h, b_eff_corr: verticalcapacity_undrained_api(
                effective_length=effective_length,
                effective_width=b_eff_corr,
                su_base=su_base,
                horizontal_load=h,
                **kwargs)['vertical_capacity [kN]'],
                                                _horizontal_load_corrected,
                                                _effective_width_corrected))
            _envelope_v_unfactored = np.append(envelope_v_corrected_raw, [0.0,])
        else:
            _envelope_v_unfactored = _envelope_v_uncorrected
    except:
        warnings.warn("Base depth not defined, assuming surface foundation", Warning)
        _envelope_v_unfactored = _envelope_v_uncorrected
    finally:
        _envelope_h_unfactored = _envelope_h_uncorrected
        _envelope_v_factored = _envelope_v_unfactored / factor_bearing
        _envelope_h_factored = _envelope_h_unfactored / factor_sliding

    return {
        'envelope_v_unfactored [kN]': _envelope_v_unfactored,
        'envelope_h_unfactored [kN]': _envelope_h_unfactored,
        'envelope_v_factored [kN]': _envelope_v_factored,
        'envelope_h_factored [kN]': _envelope_h_factored,
        'envelope_v_uncorrected [kN]': _envelope_v_uncorrected,
        'envelope_h_uncorrected [kN]': _envelope_h_uncorrected,
        'sliding_capacity': _sliding_capacity,
        'bearing_capacity': _bearing_capacity,
    }


NQ_FRICTIONANGLE_SAND = {
    'friction_angle': {'type': 'float', 'min_value': 20.0, 'max_value': 50.0},
}

NQ_FRICTIONANGLE_SAND_ERRORRETURN = {
    'Nq [-]': np.nan,
}


@Validator(NQ_FRICTIONANGLE_SAND, NQ_FRICTIONANGLE_SAND_ERRORRETURN)
def nq_frictionangle_sand(
        friction_angle,
        **kwargs):
    """
    Calculate the bearing capacity factor Nq from the friction angle

    :param friction_angle: Peak effective friction angle (:math:`\\phi_p^{\\prime}`) [:math:`deg`] - Suggested range: 20.0 <= friction_angle <= 50.0

    .. math::
        N_q = e^{\\pi \\tan \\phi_p^{\\prime}} \\tan^2 \\left( 45^{\\circ} + \\frac{ \\phi_p^{\\prime}}{2} \\right)

    :returns: Dictionary with the following keys:

        - 'Nq [-]': Bearing capacity factor (:math:`N_q`)  [:math:`-`]

    Reference - Budhu (2011) Introduction to soil mechanics and foundations

    """

    _Nq = np.exp(np.pi * np.tan(np.radians(friction_angle))) * \
          ((np.tan(np.radians(45.0 + 0.5 * friction_angle))) ** 2.0)

    return {
        'Nq [-]': _Nq,
    }


NGAMMA_FRICTIONANGLE_VESIC = {
    'friction_angle': {'type': 'float', 'min_value': 20.0, 'max_value': 50.0},
}

NGAMMA_FRICTIONANGLE_VESIC_ERRORRETURN = {
    'Ngamma [-]': np.nan,
}


@Validator(NGAMMA_FRICTIONANGLE_VESIC, NGAMMA_FRICTIONANGLE_VESIC_ERRORRETURN)
def ngamma_frictionangle_vesic(
        friction_angle,
        **kwargs):
    """
    Calculates the bearing capacity factor Ngamma according to the equation proposed by Vesic (1973). Note that alternative formulations are available.

    :param friction_angle: Peak drained friction angle (:math:`\\phi_p^{\\prime}`) [:math:`deg`] - Suggested range: 20.0 <= friction_angle <= 50.0

    .. math::
        N_{\\gamma} = 2 (N_q + 1) \\tan \\phi_p^{\\prime}

    :returns: Dictionary with the following keys:

        - 'Ngamma [-]': Bearing capacity factor (:math:`N_{\\gamma}`)  [:math:`-`]

    Reference - Budhu (2011) Introduction to soil mechanics and foundations

    """

    _Ngamma = 2.0 * (nq_frictionangle_sand(friction_angle)['Nq [-]'] + 1.0) * np.tan(np.radians(friction_angle))

    return {
        'Ngamma [-]': _Ngamma,
    }


NGAMMA_FRICTIONANGLE_MEYERHOF = {
    'friction_angle': {'type': 'float', 'min_value': 20.0, 'max_value': 50.0},
    'frictionangle_multiplier': {'type': 'float', 'min_value': None, 'max_value': None},
}

NGAMMA_FRICTIONANGLE_MEYERHOF_ERRORRETURN = {
    'Ngamma [-]': np.nan,
}


@Validator(NGAMMA_FRICTIONANGLE_MEYERHOF, NGAMMA_FRICTIONANGLE_MEYERHOF_ERRORRETURN)
def ngamma_frictionangle_meyerhof(
        friction_angle,
        frictionangle_multiplier=1.4, **kwargs):
    """
    Calculates the bearing capacity factor Ngamma according to the equation proposed by Meyerhof (1976). This formulation is more conservative compared to the Vesic formulation.

    :param friction_angle: Peak drained friction angle (:math:`\\phi_p^{\\prime}`) [:math:`deg`] - Suggested range: 20.0 <= friction_angle <= 50.0
    :param frictionangle_multiplier: Multiplier on the friction angle (:math:`\\alpha_1`) [:math:`-`] (optional, default= 1.4)

    .. math::
        N_{\\gamma} = (N_q - 1) \\tan (1.4 \\phi_p^{\\prime})

    :returns: Dictionary with the following keys:

        - 'Ngamma [-]': Bearing capacity factor (:math:`N_{\\gamma}`)  [:math:`-`]

    Reference - Budhu (2011) Introduction to soil mechanics and foundations

    """

    _Ngamma = (nq_frictionangle_sand(friction_angle)['Nq [-]'] - 1.0) * np.tan(
        np.radians(frictionangle_multiplier * friction_angle))

    return {
        'Ngamma [-]': _Ngamma,
    }


NGAMMA_FRICTIONANGLE_DAVISBOOKER = {
    'friction_angle': {'type': 'float', 'min_value': 20.0, 'max_value': 50.0},
    'roughness_factor': {'type': 'float', 'min_value': 0.0, 'max_value': 1.0},
    'multiplier_smooth': {'type': 'float', 'min_value': None, 'max_value': None},
    'multiplier_rough': {'type': 'float', 'min_value': None, 'max_value': None},
    'multiplier_exp_smooth': {'type': 'float', 'min_value': None, 'max_value': None},
    'multiplier_exp_rough': {'type': 'float', 'min_value': None, 'max_value': None},
}

NGAMMA_FRICTIONANGLE_DAVISBOOKER_ERRORRETURN = {
    'Ngamma [-]': np.nan,
    'Ngamma_smooth [-]': np.nan,
    'Ngamma_rough [-]': np.nan,
}

@Validator(NGAMMA_FRICTIONANGLE_DAVISBOOKER, NGAMMA_FRICTIONANGLE_DAVISBOOKER_ERRORRETURN)
def ngamma_frictionangle_davisbooker(
        friction_angle, roughness_factor,
        multiplier_smooth=0.0663, multiplier_rough=0.1054, multiplier_exp_smooth=9.3, multiplier_exp_rough=9.6,
        **kwargs):
    """
    Calculates the bearing capacity factor Ngamma according to the equation proposed by Davis and Booker (1971). This formulation is based on a more refined plasticity method and takes the roughness into account. This method is preferred in principle.

    :param friction_angle: Peak drained friction angle (:math:`\\phi_p^{\\prime}`) [:math:`deg`] - Suggested range: 20.0 <= friction_angle <= 50.0
    :param roughness_factor: Footing roughness factor where 0 is fully smooth and 1 is fully rough (:math:`R_{inter}`) [:math:`-`] - Suggested range: 0.0 <= roughness_factor <= 1.0
    :param multiplier_smooth: Multiplier for smooth footings (:math:`\\alpha_1`) [:math:`-`] (optional, default= 0.0663)
    :param multiplier_rough: Multiplier for rough footings (:math:`\\alpha_2`) [:math:`-`] (optional, default= 0.1054)
    :param multiplier_exp_smooth: Multiplier on exponential term for smooth footings (:math:`\\alpha_3`) [:math:`-`] (optional, default= 9.3)
    :param multiplier_exp_rough: Multiplier on exponential term for roughfootings (:math:`\\alpha_4`) [:math:`-`] (optional, default= 9.6)

    .. math::
        N_{\\gamma} = \\begin{cases}
            0.1054 \\exp (9.6 \\phi_p^{\\prime})       & \\quad \\text{for rough footings}\\\\
            0.0663 \\exp(9.3 \\phi_p^{\\prime})  & \\quad \\text{for smooth footings}
          \\end{cases}

    :returns: Dictionary with the following keys:

        - 'Ngamma [-]': Bearing capacity factor (:math:`N_{\\gamma}`)  [:math:`-`]
        - 'Ngamma_smooth [-]': Bearing capacity factor for smooth footing (:math:`N_{\\gamma,smooth}`)  [:math:`-`]
        - 'Ngamma_rough [-]': Bearing capacity factor for rough footing (:math:`N_{\\gamma,rough}`)  [:math:`-`]

    Reference - Budhu (2011) Introduction to soil mechanics and foundations

    """
    _Ngamma_smooth = multiplier_smooth * np.exp(multiplier_exp_smooth * np.radians(friction_angle))
    _Ngamma_rough = multiplier_rough * np.exp(multiplier_exp_rough * np.radians(friction_angle))
    _Ngamma = np.interp(roughness_factor, [0.0, 1.0], [_Ngamma_smooth, _Ngamma_rough])

    return {
        'Ngamma [-]': _Ngamma,
        'Ngamma_smooth [-]': _Ngamma_smooth,
        'Ngamma_rough [-]': _Ngamma_rough,
    }

class APIMudmatCalculation(object):

    def __init__(self, title):
        self.title = title

    def set_rectangular_dimensions(self, length, width, eccentricity_length, eccentricity_width):
        """
        Sets dimensions for a rectangular foundation

        :param length:
        :param width:
        :param eccentricity_length:
        :param eccentricity_width:
        :return:
        """
        pass


class APIMudmatCalculationDrained(APIMudmatCalculation):

    pass


class APIMudmatCalculationUndrained(APIMudmatCalculation):
    pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import os
import warnings

# 3rd party packages
import pandas as pd

# Project imports


def read_ags(file_path, groupname, combine_headers=True, includes_type=True):
    """
    Reads AGS data from a file and extracts the data for a given groupname into a DataFrame.
    All data with type "DP" is converted to float format

    :param file_path: Path (absolute or relative to the ags file)
    :param groupname: Name of the AGS group exactly as it is written in the AGS file
    :param combine_headers: Boolean determining whether the units are included in the header or not
    :param includes_type: Boolean determining whether a TYPE is included with the data
    :return: Dataframe with data for the given group
    """

    try:
        with open(file_path) as f:
            # Read the file as raw data
            ags_raw = pd.DataFrame(f.readlines(), columns=["AGS lines"])
    except Exception as err:
        raise IOError("Error during reading of file %s - %s" % (file_path, str(err)))

    # Find the index where the data for the given groupname starts
    try:
        if includes_type:
            start_index = ags_raw[ags_raw['AGS lines'].str.startswith(r'"GROUP","%s"' % groupname)].index[0]
        else:
            start_index = ags_raw[ags_raw['AGS lines'].str.startswith(r'"%s"' % groupname)].index[0]
    except:
        raise ValueError("Groupname not found in AGS file")

    try:
        # Find the index of the row where the data ends
        empty_rows = ags_raw[ags_raw['AGS lines'] == "\n"]
        if empty_rows.__len__() == 0:
            raise ValueError("AGS file not correctly structured, no, empty rows between groups")
        end_index = empty_rows[empty_rows.index > start_index].index[0]
    except:
        warnings.warn("No empty row detected below group, reading to the end of the file")
        end_index = ags_raw.index[-1]

    # Read the data as csv
    try:
        group_data = pd.read_csv(file_path, skiprows=start_index + 1, nrows=end_index - start_index-2)
    except Exception as err:
        raise ValueError("Error during reading of group data - %s" % str(err))

    # Convert column names and datatypes of numerical columns
    try:
        if includes_type:
            new_headers = []
            datatypes = dict()
            for i, original_header in enumerate(group_data.columns):
                if combine_headers:
                    new_name = "%s [%s]" % (original_header, group_data.loc[0, original_header])
                else:
                    new_name = original_header
                new_headers.append(new_name)
                if "DP" in group_data.loc[1, original_header]:
                    datatypes[new_name] = 'float'
            group_data.columns = new_headers
            group_data = group_data[2:].reset_index(drop=True).astype(datatypes)
        else:
            group_data = group_data[1:].reset_index(drop=True)
    except Exception as err:
        raise ValueError(
            "Error during the modification of column headers, check AGS file format - %s" % str(err))

    return group_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages

# 3rd party packages
import numpy as np
from scipy.integrate import cumtrapz

# Project imports
from groundhog.general.validation import Validator


NEGATIVESKINFRICTION_PILEGROUP_ZEEVAERTDEBEER = {
    'depths': {'type': 'list', 'elementtype': 'float', 'order': 'ascending', 'unique': True, 'empty_allowed': False},
    'effective_unit_weights': {'type': 'list', 'elementtype': 'float', 'order': None, 'unique': False,
                               'empty_allowed': False},
    'lateral_earth_pressure_coefficients': {'type': 'list', 'elementtype': 'float', 'order': None, 'unique': False,
                                            'empty_allowed': False},
    'interface_friction_angles': {'type': 'list', 'elementtype': 'float', 'order': None, 'unique': False,
                                  'empty_allowed': False},
    'surcharge': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'diameter': {'type': 'float', 'min_value': 0.01, 'max_value': 10.0},
    'diameter_influence': {'type': 'float', 'min_value': 0.0, 'max_value': None},
}

NEGATIVESKINFRICTION_PILEGROUP_ZEEVAERTDEBEER_ERRORRETURN = {
    'virgin_effective_stress [kPa]': np.nan,
    'group_effective_stress [kPa]': np.nan,
    'negative_skin_friction_profile_single [kN]': np.nan,
    'negative_skin_friction_profile_group [kN]': np.nan,
    'negative_skin_friction [kN]': np.nan,
    'negative_skin_friction_group [kN]': np.nan,
}


@Validator(NEGATIVESKINFRICTION_PILEGROUP_ZEEVAERTDEBEER, NEGATIVESKINFRICTION_PILEGROUP_ZEEVAERTDEBEER_ERRORRETURN)
def negativeskinfriction_pilegroup_zeevaertdebeer(
        depths, effective_unit_weights, lateral_earth_pressure_coefficients, interface_friction_angles, surcharge,
        diameter, diameter_influence,
        **kwargs):
    """
    Calculates the negative skin friction for a pile in a pile group according to the method of Zeevaert en De Beer. To allow any unit weight profile, the differential equation is solved in finite difference form.

    :param depths: Array with depths used for the calculation (:math:`z`) [:math:`m`] - Elementtype: float, order: ascending, unique: True, empty entries allowed: False
    :param effective_unit_weights: Array with effective unit weights used for the calculation (:math:`\\gamma^{\\prime}`) [:math:`kN/m3`] - Elementtype: float, order: None, unique: False, empty entries allowed: False
    :param lateral_earth_pressure_coefficients: Array with lateral earth pressure coefficient at each depth (:math:`K_0`) [:math:`-`] - Elementtype: float, order: None, unique: False, empty entries allowed: False
    :param interface_friction_angles: Array with interface friction angles (:math:`\\delta^{\\prime}`) [:math:`deg`] - Elementtype: float, order: None, unique: False, empty entries allowed: False
    :param surcharge: Amount of stress applied on top of the soil mass (:math:`p_0^{\\prime}`) [:math:`kPa`] - Suggested range: surcharge >= 0.0
    :param diameter: Pile diameter (:math:`D_p`) [:math:`m`] - Suggested range: 0.01 <= diameter <= 10.0
    :param diameter_influence: Diameter of the zone of influence for negative skin friction (:math:`D_n`) [:math:`m`] - Suggested range: diameter_influence >= 0.0

    .. math::
        \\frac{d \\sigma_v^{\\prime}}{dz} = \\gamma^{\\prime} - \\tau \\cdot \\frac{O_s}{A}

        m = K_0 \\cdot \\tan \\delta^{\\prime} \\cdot \\frac{O_s}{A}

        \\frac{\\Delta \\sigma_{v,i+1}^{\\prime}}{\\Delta z} - \\gamma^{\\prime} + m \\cdot \\sigma_{v,i}^{\\prime} = 0

        \\sigma_{v,0}^{\\prime} = p_0^{\\prime}

        O_s = \\pi \\cdot D_p

        A = \\frac{\\pi}{4} \\cdot \\left( D_n^2 - D_p^2 \\right)

    :returns: Dictionary with the following keys:

        - 'virgin_effective_stress [kPa]': Effective stress profile in the absence of surcharge (:math:`\\sigma_{vo}^{\\prime}`)  [:math:`kPa`]
        - 'group_effective_stress [kPa]': Effective stress accounting for the effect of soil hanging on the pile (:math:`\\sigma_{v}^{\\prime}`)  [:math:`kPa`]
        - 'negative_skin_friction_profile_single [kN]': Cumulative negative skin friction for a single pile (:math:`F_{n}`)  [:math:`kN`]
        - 'negative_skin_friction_profile_group [kN]': Cumulative negative skin friction for pile in the group (:math:`F_{n, group}`)  [:math:`kN`]
        - 'negative_skin_friction [kN]': Total value of negative skin friction for a single pile (:math:`F_{n,tot}`)  [:math:`kN`]
        - 'negative_skin_friction_group [kN]': Total value of negative skin friction for a pile in a group (:math:`F_{n,tot,group}`)  [:math:`kN`]

    .. figure:: images/negativeskinfriction_pilegroup_zeevaertdebeer_1.png
        :figwidth: 500.0
        :width: 450.0
        :align: center

        Pile group with contributing soil and equilibrium of infinitesimal soil slice

    Reference - Zeevaert - De Beer (1966)

    """
    # Validation
    if diameter_influence < diameter:
        raise ValueError("Diameter of the zone of influence should be greated than the pile diameter")
    if len(depths) != len(effective_unit_weights) or len(depths) != len(lateral_earth_pressure_coefficients) or \
        len(depths) != len(interface_friction_angles):
        raise ValueError("All array inputs should have equal lengths")

    area = 0.25 * np.pi * ((diameter_influence) ** 2 - diameter ** 2)
    circumference = np.pi * diameter

    m_values = np.array(lateral_earth_pressure_coefficients) * \
               np.tan(np.radians(np.array(interface_friction_angles))) * \
               (circumference / area)

    _virgin_effective_stress = np.append(0, cumtrapz(
        y=np.array(effective_unit_weights),
        x=np.array(depths))) + surcharge

    sigma_v_neg_fd = np.zeros(len(depths))
    sigma_v_neg_fd[0] = surcharge
    for i, z in enumerate(depths):
        if i > 0:
            sigma_v_neg_fd[i] = np.diff(depths)[0] * (
                    effective_unit_weights[i] -
                    m_values[i] * sigma_v_neg_fd[i - 1] +
                    (sigma_v_neg_fd[i - 1] / np.diff(depths)[0]))
    _group_effective_stress = sigma_v_neg_fd
    _negative_skin_friction_profile_single = \
        np.append(0, cumtrapz(_virgin_effective_stress, depths)) * \
        np.pi * diameter * np.array(lateral_earth_pressure_coefficients) * \
        np.tan(np.radians(np.array(interface_friction_angles)))
    _negative_skin_friction_profile_group = \
        np.append(0, cumtrapz(sigma_v_neg_fd, depths)) * \
        np.pi * diameter * np.array(lateral_earth_pressure_coefficients) * \
        np.tan(np.radians(np.array(interface_friction_angles)))

    _negative_skin_friction = _negative_skin_friction_profile_single[-1]
    _negative_skin_friction_group = _negative_skin_friction_profile_group[-1]

    return {
        'virgin_effective_stress [kPa]': _virgin_effective_stress,
        'group_effective_stress [kPa]': _group_effective_stress,
        'negative_skin_friction_profile_single [kN]': _negative_skin_friction_profile_single,
        'negative_skin_friction_profile_group [kN]': _negative_skin_friction_profile_group,
        'negative_skin_friction [kN]': _negative_skin_friction,
        'negative_skin_friction_group [kN]': _negative_skin_friction_group,
    }


#!/usr/bin/env python3

import os
from setuptools import setup, find_packages
from groundhog.__version__ import __version__, __description__, __url__, __download_url__

# load the README file and use it as the long_description for PyPI
def readme():
    with open('README.md', 'r') as f:
        return f.read()

# package configuration - for reference see:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#id9
setup(
      name='groundhog',
      version=__version__,
      description=__description__,
      long_description=readme(),
      url=__url__,
      download_url=__download_url__,
      keywords=['engineering', 'geotechnical'],
      author='Bruno Stuyts',
      author_email='bruno@pro-found.be',
      license='Creative Commons BY-SA 4.0',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='nose.collector',
      tests_require=['nose'],
      classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
      ],
      python_requires='>=3.6',
      )


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import unittest
import os

# 3rd party packages
import pandas as pd
import numpy as np

# Project imports
from groundhog.general import validation


class Test_check_layer_overlap(unittest.TestCase):

    def setUp(self):
        self.df_no_overlap = pd.DataFrame({
            'z from [m]': np.array([0, 4, 10]),
            'z to [m]': np.array([4, 10, 20])
        })
        self.df_gaps = pd.DataFrame({
            'z from [m]': np.array([0, 5, 10]),
            'z to [m]': np.array([4, 10, 20])
        })
        self.df_overlap = pd.DataFrame({
            'z from [m]': np.array([0, 4, 10]),
            'z to [m]': np.array([5, 10, 20])
        })
        self.df_otherkeys = pd.DataFrame({
            'Depth from [m]': np.array([0, 4, 10]),
            'Depth to [m]': np.array([4, 10, 20])
        })

    def test_check_layer_overlap(self):
        validation.check_layer_overlap(self.df_no_overlap)
        validation.check_layer_overlap(self.df_otherkeys, z_from_key="Depth from [m]", z_to_key="Depth to [m]")
        # An error should be raised
        self.assertRaises(ValueError, validation.check_layer_overlap, df=self.df_gaps)
        self.assertRaises(ValueError, validation.check_layer_overlap, df=self.df_overlap)
        validation.check_layer_overlap(self.df_gaps, raise_error=False)
        validation.check_layer_overlap(self.df_overlap, raise_error=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import unittest
import os

# 3rd party packages
import pandas as pd
import numpy as np

# Project imports
from groundhog.general.validation import Validator


VOIDRATIO_POROSITY = {
    'porosity': {'type': 'float', 'min_value': 0.0, 'max_value': 1.0},
}

VOIDRATIO_POROSITY_ERRORRETURN = {
    'voidratio [-]': np.nan,
}

@Validator(VOIDRATIO_POROSITY, VOIDRATIO_POROSITY_ERRORRETURN)
def voidratio_porosity(
        porosity,
        **kwargs):

    """
    Converts a void ratio into a porosity

    :param porosity: Porosity of the sample defined as the ratio of volume of voids to total volume (:math:`n`) [:math:`-`] - Suggested range: 0.0 <= porosity <= 1.0

    .. math::
        e = \\frac{V_{voids}}{V_{solids}}

        n = \\frac{V_{voids}}{V_{total}}

        e = \\frac{n}{1-n}

    :returns: Dictionary with the following keys:

        - 'voidratio [-]': Void ratio defined as the ratio of volume of voids to volume of solids (:math:`e`)  [:math:`-`]

    Reference - Budhu (2011). Soil mechanics and foundation engineering

    """

    _voidratio = porosity / (1- porosity)

    return {
        'voidratio [-]': _voidratio,
    }


POROSITY_CALC = {
    'voidratio': {'type': 'float', 'min_value': 0.0, 'max_value': 5.0},
}

POROSITY_CALC_ERRORRETURN = {
    'porosity [-]': np.nan,
}


@Validator(POROSITY_CALC, POROSITY_CALC_ERRORRETURN)
def porosity_voidratio(
        voidratio,
        **kwargs):
    """
    Calculates the porosity of sample from the void ratio

    :param voidratio: Void ratio defined as the ratio of volume of voids to volume of solids (:math:`e`) [:math:`-`] - Suggested range: 0.0 <= voidratio <= 5.0

    .. math::
        n = \\frac{e}{e+1}

    :returns: Dictionary with the following keys:

        - 'porosity [-]': Porosity defined as the ratio of the volume of voids to the total volume (:math:`n`)  [:math:`-`]

    Reference - Budhu (2011). Soil mechanics and foundation engineering

    """

    _porosity = voidratio / (1 + voidratio)

    return {
        'porosity [-]': _porosity,
    }


SATURATION_WATERCONTENT = {
    'water_content': {'type': 'float', 'min_value': 0.0, 'max_value': 4.0},
    'voidratio': {'type': 'float', 'min_value': 0.0, 'max_value': 4.0},
    'specific_gravity': {'type': 'float', 'min_value': 1.0, 'max_value': 3.0},
}

SATURATION_WATERCONTENT_ERRORRETURN = {
    'saturation [-]': np.nan,
}


@Validator(SATURATION_WATERCONTENT, SATURATION_WATERCONTENT_ERRORRETURN)
def saturation_watercontent(
        water_content, voidratio,
        specific_gravity=2.65, **kwargs):
    """
    Calculates the saturation of a sample from the water content, the specific gravity and the void ratio

    :param water_content: Water content of the soil defined as the ratio of weight of water to weight of solids (:math:`w`) [:math:`-`] - Suggested range: 0.0 <= water_content <= 4.0
    :param voidratio: Ratio of volume of voids to volume of solids (:math:`e`) [:math:`-`] - Suggested range: 0.0 <= voidratio <= 4.0
    :param specific_gravity: Specific gravity of the soil grains (:math:`G_s`) [:math:`-`] - Suggested range: 1.0 <= specific_gravity <= 3.0 (optional, default= 2.65)

    .. math::
        S = \\frac{V_{water}}{V_{voids}} = \\frac{w \\cdot G_s}{e}

    :returns: Dictionary with the following keys:

        - 'saturation [-]': Saturation of the sample defined as the ratio of volume of water to volume of voids (:math:`S`)  [:math:`-`]

    Reference - Budhu (2011). Soil mechanics and foundation engineering

    """

    _saturation = water_content * specific_gravity / voidratio

    return {
        'saturation [-]': _saturation,
    }


BULKUNITWEIGHT = {
    'saturation': {'type': 'float', 'min_value': 0.0, 'max_value': 1.0},
    'voidratio': {'type': 'float', 'min_value': 0.0, 'max_value': 4.0},
    'specific_gravity': {'type': 'float', 'min_value': 1.0, 'max_value': 3.0},
    'unitweight_water': {'type': 'float', 'min_value': 9.0, 'max_value': 11.0},
}

BULKUNITWEIGHT_ERRORRETURN = {
    'bulk unit weight [kN/m3]': np.nan,
    'effective unit weight [kN/m3]': np.nan,
}


@Validator(BULKUNITWEIGHT, BULKUNITWEIGHT_ERRORRETURN)
def bulkunitweight(
        saturation, voidratio,
        specific_gravity=2.65, unitweight_water=10.0, **kwargs):
    """
    Calculates the bulk unit weight from specific gravity, void ratio and saturation

    :param saturation: Saturation of the sample, ratio of volume of water to volume of voids (:math:`S`) [:math:`-`] - Suggested range: 0.0 <= saturation <= 1.0
    :param voidratio: Void ratio, ratio of volume of voids to volume of solids (:math:`e`) [:math:`-`] - Suggested range: 0.0 <= voidratio <= 4.0
    :param specific_gravity: Specific gravity of solid particles (:math:`G_s`) [:math:`-`] - Suggested range: 1.0 <= specific_gravity <= 3.0 (optional, default= 2.65)
    :param unitweight_water: Unit weight of water (:math:`\\gamma_w`) [:math:`kN/m3`] - Suggested range: 9.0 <= unitweight_water <= 11.0 (optional, default= 10.0)

    .. math::
        \\gamma = \\frac{W}{V} = \\left( \\frac{G_s + S \\cdot e}{1+e} \\right) \\cdot \\gamma_w

    :returns: Dictionary with the following keys:

        - 'bulk unit weight [kN/m3]': Bulk unit weight of the material (:math:`\\gamma`)  [:math:`kN/m3`]
        - 'effective unit weight [kN/m3]': Effective unit weight of the material (:math:`\\gamma^{\\prime}`)  [:math:`kN/m3`]

    Reference - Budhu (2011). Soil mechanics and foundation engineering

    """

    _bulk_unit_weight = unitweight_water * (
            (specific_gravity + saturation * voidratio) / (1 + voidratio))
    _effective_unit_weight = _bulk_unit_weight - unitweight_water

    return {
        'bulk unit weight [kN/m3]': _bulk_unit_weight,
        'effective unit weight [kN/m3]': _effective_unit_weight,
    }


DRYUNITWEIGHT_WATERCONTENT = {
    'watercontent': {'type': 'float', 'min_value': 0.0, 'max_value': 4.0},
    'bulkunitweight': {'type': 'float', 'min_value': 10.0, 'max_value': 25.0},
}

DRYUNITWEIGHT_WATERCONTENT_ERRORRETURN = {
    'dry unit weight [kN/m3]': np.nan,
}


@Validator(DRYUNITWEIGHT_WATERCONTENT, DRYUNITWEIGHT_WATERCONTENT_ERRORRETURN)
def dryunitweight_watercontent(
        watercontent, bulkunitweight,
        **kwargs):
    """
    Calculates the dry unit weight of the sample from the water content and the bulk unit weight

    :param watercontent: Water content of the sample, ratio of weight of water to weight of solids (:math:`w`) [:math:`-`] - Suggested range: 0.0 <= watercontent <= 4.0
    :param bulkunitweight: Bulk unit weight of the sample (:math:`\\gamma`) [:math:`kN/m3`] - Suggested range: 10.0 <= bulkunitweight <= 25.0

    .. math::
        \\gamma_d = \\frac{W_s}{V} = \\left( \\frac{G_s}{1+e} \\right) \\cdot \\gamma_w = \\frac{\\gamma}{1 + w}

    :returns: Dictionary with the following keys:

        - 'dry unit weight [kN/m3]': Dry unit weight, ratio of weight of solids to total volume (:math:`\\gamma_d`)  [:math:`kN/m3`]

    Reference - Budhu (2011). Soil mechanics and foundation engineering

    """

    _dry_unit_weight = bulkunitweight / (1 + watercontent)

    return {
        'dry unit weight [kN/m3]': _dry_unit_weight,
    }


BULKUNITWEIGHT_DRYUNITWEIGHT = {
    'dryunitweight': {'type': 'float', 'min_value': 1.0, 'max_value': 15.0},
    'watercontent': {'type': 'float', 'min_value': 0.0, 'max_value': 4.0},
}

BULKUNITWEIGHT_DRYUNITWEIGHT_ERRORRETURN = {
    'bulk unit weight [kN/m3]': np.nan,
    'effective unit weight [kN/m3]': np.nan,
}


@Validator(BULKUNITWEIGHT_DRYUNITWEIGHT, BULKUNITWEIGHT_DRYUNITWEIGHT_ERRORRETURN)
def bulkunitweight_dryunitweight(
        dryunitweight, watercontent, unitweight_water=10.0,
        **kwargs):
    """
    Calculates the bulk unit weight from the dry unit weight and the water content

    :param dryunitweight: Dry unit weight, ratio of weight of solids to total volume (:math:`\\gamma_d`) [:math:`kN/m3`] - Suggested range: 1.0 <= dryunitweight <= 15.0
    :param watercontent: Water content, ratio of weight of water to weight of solids (:math:`w`) [:math:`-`] - Suggested range: 0.0 <= watercontent <= 4.0
    :param unitweight_water: Unit weight of water (:math:`\\gamma_w`) [:math:`kN/m3`] - Suggested range: 9.0 <= unitweight_water <= 11.0 (optional, default= 10.0)

    .. math::
        \\gamma = (1+w) \\cdot \\gamma_d

    :returns: Dictionary with the following keys:

        - 'bulk unit weight [kN/m3]': Bulk unit weight (:math:`\\gamma`)  [:math:`kN/m3`]
        - 'effective unit weight [kN/m3]': Effective unit weight (:math:`\\gamma^{\\prime}`)  [:math:`kN/m3`]

    Reference - Budhu (2011). Soil mechanics and foundation engineering

    """

    _bulk_unit_weight = (1 + watercontent) * dryunitweight
    _effective_unit_weight = _bulk_unit_weight - unitweight_water

    return {
        'bulk unit weight [kN/m3]': _bulk_unit_weight,
        'effective unit weight [kN/m3]': _effective_unit_weight,
    }


RELATIVE_DENSITY = {
    'void_ratio': {'type': 'float', 'min_value': 0.0, 'max_value': 5.0},
    'e_min': {'type': 'float', 'min_value': 0.0, 'max_value': 5.0},
    'e_max': {'type': 'float', 'min_value': 0.0, 'max_value': 5.0},
}

RELATIVE_DENSITY_ERRORRETURN = {
    'Dr [-]': np.nan,
}


@Validator(RELATIVE_DENSITY, RELATIVE_DENSITY_ERRORRETURN)
def relative_density(void_ratio, e_min, e_max, **kwargs):
    """
    Calculates the relative density for a cohesionless sample from the measured void ratio, comparing it to the void ratio at minimum and maximum density.

    :param void_ratio: Void ratio of the sample (:math:`e`) [:math:`-`] - Suggested range: 0.0 <= void_ratio <= 5.0
    :param e_min: Void ratio at the minimum density (:math:`e_{min}`) [:math:`-`] - Suggested range: 0.0 <= e_min <= 5.0
    :param e_max: Void ratio at the maximum density (:math:`e_{max}`) [:math:`-`] - Suggested range: 0.0 <= e_max <= 5.0

    .. math::
        D_r = \\frac{e - e_{min}}{e_{max} - e_{min}}

    :returns: Dictionary with the following keys:

        - 'Dr [-]': Relative density (:math:`D_r`)  [:math:`-`]

    Reference - Budhu (2011). Soil mechanics and foundation engineering

    """

    _Dr = (void_ratio - e_min) / (e_max - e_min)

    return {
        'Dr [-]': _Dr,
    }

RELATIVEDENSITY_CATEGORIES = {
    'relative_density': {'type': 'float', 'min_value': 0.0, 'max_value': 1.0},
}

RELATIVEDENSITY_CATEGORIES_ERRORRETURN = {
    'Relative density []': None,
}


@Validator(RELATIVEDENSITY_CATEGORIES, RELATIVEDENSITY_CATEGORIES_ERRORRETURN)
def relativedensity_categories(
        relative_density,
        **kwargs):
    """
    Categorizes relative densities according to the following definition:

        - 0 - 0.15: Very loose
        - 0.15 - 0.35: Loose
        - 0.35 - 0.65: Medium dense
        - 0.65 - 0.85: Dense
        - 0.85 - 1: Very dense

    :param relative_density: Relative density of cohesionless material (:math:`D_r`) [:math:`-`] - Suggested range: 0.0 <= relative_density <= 1.0

    .. math::
        D_r = \\frac{e - e_{min}}{e_{max} - e_{min}}

    :returns: Dictionary with the following keys:

        - 'Relative density': Relative density class

    Reference - API RP2 GEO

    """

    if 0 <= relative_density < 0.15:
        _relative_density = "Very loose"
    elif 0.15 <= relative_density < 0.35:
        _relative_density = "Loose"
    elif 0.35 <= relative_density < 0.65:
        _relative_density = "Medium dense"
    elif 0.65 <= relative_density < 0.85:
        _relative_density = "Dense"
    elif relative_density > 0.85:
        _relative_density = "Very dense"

    return {
        'Relative density': _relative_density,
    }


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import warnings
from copy import deepcopy

# 3rd party packages
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from scipy.optimize import brentq
try:
    from plotly import tools, subplots
    import plotly.graph_objs as go
    from plotly.colors import DEFAULT_PLOTLY_COLORS
    from plotly.offline import iplot
except:
    warnings.warn('Plotly could not be imported. Install plotly using conda (conda install plotly) to enable interactive plotting of results')

# Project imports


class DeBeerCalculation(object):

    def __init__(self, depth, qc, diameter_pile, diameter_cone=0.0357):
        """
        Initializes a pile base resistance calculation according to De Beer's method. Lists or Numpy arrays of depth
        and cone resistance need to be supplied to the routine as well as the diameter of the pile and the cone.
        The depth spacing does not need to be equal to the spacing required by De Beer's method. The `resample_data`
        method is used after calculation initialisation to remap the data onto a 0.2m grid.

        :param depth: List or Numpy array with depths [m]
        :param qc: List or Numpy array with cone resistance values [MPa] - same list length as depth array
        :param diameter_pile: Diameter of the pile [m]
        :param diameter_cone: Diameter of the cone, the cone diameter is 0.0357 for a standard cone [m]
        """

        # Validation

        if depth.__len__() != qc.__len__():
            raise ValueError("depth and qc arrays need to have the same length!")
        if diameter_pile < 0.2:
            raise ValueError("The minimum pile diameter for applying De Beer's method is 0.2m")
        self.depth_raw = np.array(depth)
        self.qc_raw = np.array(qc)
        self.diameter_pile = diameter_pile
        self.diameter_cone = diameter_cone

    def resample_data(self, spacing=0.2):
        """
        Resampling of the data to the required spacing (default=0.2m for mechanical cone) can be performed for e.g.
        piezocone data.

        :param spacing: Spacing used for calculation of De Beer's method
        :return:
        """
        self.h_crit = spacing
        self.depth = np.arange(np.array(self.depth_raw).min(), np.array(self.depth_raw).max(), spacing)
        self.qc_mech = np.interp(self.depth, self.depth_raw, self.qc_raw)

    def set_soil_layers(self, soilprofile, soiltypecolumn="Soil type",
                        tertiaryclaycolumn="Tertiary clay",
                        totalunitweightcolumn="Total unit weight [kN/m3]",
                        water_level=0, total_unit_weight_dry=15.696,
                        total_unit_weight_wet=19.62, **kwargs):
        """
        Sets the soil type for the pile resistance calculation.
        For the shaft resistance calculation, a column 'Soil type' is required.

        The calculation of overburden pressure is included in this routine.
        According to the Belgian practice, a total unit weight of 15.696:math:`kN/m^3` is used for dry soil and
        a unit weight of 19.62:math:`kN/m^3` for wet soil. An array with total unit weights for each soil layer
        can be specified (`total_unit_weight`) to override these defaults. If the water level is not at a layer
        interface, this array need to contain an additional entry to account for the difference between the dry and wet
        unit weight in the layer containing the water table.

        The unit weight used in De Beer's base
        resistance calculation is the total unit weight above the water table
        and the effective unit weight below.

        The soil types need to be specified in accordance with Table 4 of the paper
        Design of deepfoundations - Belgian practice (Huybrechts et al, 2016).

        Finally, it is possible to specify whether each layer is tertiary clay or not using an array of Booleans.
        Occurence of stiff tertiary clays can be taken into account by specifying the array `tertiary_clay`.

        :param soilprofile: SoilProfile object containing the layer definitions. The SoilProfile should con
        :param soiltypecolumn: Name of the column containing the soil type, select from "Clay", "Loam (silt)", "Sandy clay / loam (silt)", "Clayey sand / loam (silt)" and "Sand"
        :param tertiaryclaycolumn: Name of the column containing the booleans determining whether the layer is Tertiary clay or not
        :param water_level: Water level used for the effective stress calculation [m], default = 0 for water level at surface
        :param water_unit_weight: Unit weight of water used for the calculations (default=10:math:`kN/m^3`)
        :param total_unit_weight_dry: Dry unit weight used for all soils (default=15.696:math:`kN/m^3`)
        :param total_unit_weight_wet: Wet unit weight used for all soils (default=19.62:math:`kN/m^3`)
        :param total_unit_weight: Array with total unit weight for each soil layer. Specifying this array will override the defaults (default=`None`)
        """

        self.layering = soilprofile

        # Validate the presence of the column with the Soil Type
        if soiltypecolumn not in soilprofile.string_soil_parameters():
            raise ValueError("SoilProfile does not contain a column with the soil type")

        for i, layer in soilprofile.iterrows():
            if layer[soiltypecolumn] not in ["Clay", "Loam (silt)", "Sandy clay / loam (silt)",
                                  "Clayey sand / loam (silt)", "Sand"]:
                raise ValueError("Soil type %s not recognised. Needs to be one of 'Loam (silt)', "
                                 "'Sandy clay / loam (silt)', 'Clayey sand / loam (silt)' or 'Sand'")

        # Validate the extent of the soil profile
        if soilprofile.min_depth != 0:
            raise ValueError("Layering should start from zero depth")
        if soilprofile.max_depth < self.depth.max():
            raise ValueError("Layering should be defined to the bottom of the CPT")

        # Validation of specified water level
        if water_level < 0:
            raise ValueError("Specified water level should be greater than or equal to zero.")

        # Set whether soil layers are tertiary clay or not
        if tertiaryclaycolumn not in soilprofile.string_soil_parameters():
            self.layering.loc[:, "Tertiary clay"] = False
        else:
            pass

        # Add a layer interface for the water level
        if water_level not in self.layering.layer_transitions():
            self.layering.insert_layer_transition(depth=water_level)

        # Check if a total unit weight has been specified. If not, use the defaults from Belgian practice
        if totalunitweightcolumn not in soilprofile.numerical_soil_parameters():
            for i, layer in self.layering.iterrows():
                if 0.5 * (layer[self.layering.depth_from_col] + layer[self.layering.depth_to_col]) < water_level:
                    self.layering.loc[i, totalunitweightcolumn] = total_unit_weight_dry
                else:
                    self.layering.loc[i, totalunitweightcolumn] = total_unit_weight_wet
        else:
            pass

        # Calculate stresses in the layers
        self.layering.calculate_overburden(
            waterlevel=water_level,
            totalunitweightcolumn=totalunitweightcolumn
        )

        self.calculation_data = self.layering.map_soilprofile(nodalcoords=self.depth)
        self.calculation_data['qc [MPa]'] = self.qc_mech

    @staticmethod
    def phi_func():
        """
        This function implements the Equations 22 and 23 of the original paper by De Beer for the calculation of
        the friction angle from the cone tip resistance and the vertical effective stress level
        :return: An interpolation function providing the friction angle as a function of the ratio of cone tip
        resistance to vertical effective stress.

        .. math::
            V_{b,d} = \\frac{q_c}{p_o}

            V_{b,d} = 1.3 \\cdot \\exp \\left( 2 \\cot \\pi \\cdot \\tan \\varphi \\right) \\tan ^2 \\left( \\frac{\\pi}{4} + \\frac{\\varphi}{2} \\right)
        """
        phi = np.linspace(np.deg2rad(0.01), np.deg2rad(50), 250)

        def v_bd_func(frictionangle):
            return 1.3 * np.exp(2 * np.pi * np.tan(frictionangle)) * ((np.tan(0.25 * np.pi + 0.5 * frictionangle)) ** 2)

        v_bd = v_bd_func(phi)

        return interp1d(v_bd, phi)

    @staticmethod
    def optimisation_func(beta, hd, frictionangle):
        """
        This function is used to find the angle beta for the failure mechanism for CPT and pile.
        The function implements Equation 60 of the original paper by De Beer.
        To obtain the value of beta, the result of evaluating Equation 60 is compared to the actual h/d or h/D
        and the root is found.

        .. math::
            \\left( \\frac{h}{d} \\right)_3 = \\frac{\\tan \\left( \\frac{\\pi}{4} + \\frac{\\varphi}{2} \\right) \\cdot \\exp \\left( \\frac{\\pi}{2} \\cdot \\tan \\varphi \\right) \\cdot \\sin \\left( \\beta \\cdot \\exp(\\beta \\cdot \\tan \\varphi) \\right)}{1 + \\delta \\cdot \\sin(2 \\cdot \\varphi)}

        :param beta: Value of beta to be found
        :param hd: Value of h/D for the pile or h/d for the CPT
        :param frictionangle: Friction angle derived from Equations 22 and 23 of De Beer's paper
        :return: Difference of the calculated and specified value of h/D or h/d
        """
        return ((np.tan(0.25 * np.pi + 0.5 * frictionangle) *
                np.exp(0.5 * np.pi * np.tan(frictionangle)) *
                np.sin(beta) *
                np.exp(beta * np.tan(frictionangle))) / \
               (1 + np.sin(2 * frictionangle))) - hd

    @staticmethod
    def stress_correction(qc, po, diameter_pile, diameter_cone, gamma, hcrit=0.2):
        """
        Calculates the stress correction in De Beer's method

        .. math::
            h^{\\prime}_{crit} = h_{crit} \\cdot D / d

            q_{r,crit} = \\frac{1 + \\frac{\\gamma \\cdot h_{crit}^{\\prime}}{2 \\cdot p_o}}{1 + \\frac{\\gamma \\cdot h_{crit}}{2 \\cdot p_o}} \\cdot q_{c,crit}

        :param qc: Cone tip resistance [MPa]
        :param po: Overburden pressure at the depth of the start of the increase [kPa]
        :param diameter_pile: Diameter of the pile [m]
        :param diameter_cone: Diameter of the cone rod [m]
        :param gamma: Unit weight of the soil (total above the water table, effective below) [kN/m3]
        :param hcrit: Distance to cone tip used in De Beer's method (default=0.2) [m]

        :returns Ultimate bearing resistance for the cone [MPa]
        """
        h_prime_crit = hcrit * (diameter_pile / diameter_cone)
        return qc * ((1.0 + ((gamma * h_prime_crit) / (2.0 * po))) / (1.0 + ((gamma * hcrit) / (2.0 * po))))

    def calculate_base_resistance(self, vanimpecorrection=False, hcrit=0.2):
        """
        Calculates the base resistance for any pile diameter.

        If the pile diameter is not a multiple of 0.2m, the unit base resistance for diameters which are a
        multiple of 0.2m above and below the given diameter are calculated.
        Linear interpolation is then used to obtain the end result

        :param vanimpecorrection: Boolean determining whether the upward correction according to De Beer's paper (default) or Van Impe (multiplier of 2) needs to be applied.
        :param hcrit: :math:`h_{crit}` adopted for De Beer's calculation (based on the mechanical cone). Default=0.2m
        :return: Sets two dataframes `calc_1` and `calc_2`, one for the multiple of 0.2m lower than the given diameter and one for the multiple of 0.2m higher than the given diameter. Finally, the attribute `qb` is set through linear interpolation.
        """

        if self.diameter_pile % 0.2 == 0:
            # Calculation for pile diameter being a multiple of 0.2m
            self.diameter_1 = self.diameter_pile
            self.diameter_2 = self.diameter_pile
        else:
            self.diameter_1 = np.round(self.diameter_pile - (self.diameter_pile % 0.2), 1)
            self.diameter_2 = np.round(self.diameter_pile + (self.diameter_pile % 0.2), 1)
        self.calc_1 = self.calculate_base_resistance_standard_diameter(
            pile_diameter=self.diameter_1, vanimpecorrection=vanimpecorrection, hcrit=hcrit
        )
        self.calc_2 = self.calculate_base_resistance_standard_diameter(
            pile_diameter=self.diameter_2, vanimpecorrection=vanimpecorrection, hcrit=hcrit
        )
        _qb = []
        for i, row in self.calc_1.iterrows():
            _qb.append(
                np.interp(
                    self.diameter_pile,
                    [self.diameter_1, self.diameter_2],
                    [self.calc_1.loc[i, "qb [MPa]"], self.calc_2.loc[i, "qb [MPa]"]]
                )
            )
        self.qb = np.array(_qb)
        self.depth_qb = self.calc_1['z [m]']

    def calculate_base_resistance_standard_diameter(self, pile_diameter, vanimpecorrection=False, hcrit=0.2):
        """
        Calculates the base resistance according to De Beer's method for a pile diameter which is a multiple of 0.2m. The calculation happens in five steps:

            - Step 1: Correct the cone resistance for the different failure surface for a pile and a CPT using Equation 62 from De Beer's paper. This correction is especially necessary for the shallow layer where the angle beta is lower than 90°
            - Step 2: Apply a correction for the different stress level for a pile compared to a CPT
            - Step 3: Account for the transition from weaker the stronger layers by working downward along the CPT trace. The increase of resistance will be slower for a pile compared to a CPT
            - Step 4: Account for the transition from stronger to weaker layers by working through the CPT trace from the bottom up. A weaker layer will be felt sooner by the model pile than by the CPT
            - Step 5: Take the average unit base resistance for one diameter below the given level. The average value should note be greater than :math:`q_{p,q+1}` at the given level.

        .. math::
            \\text{Step 1}

            q_{p,(1)} = \\frac{q_c}{\\exp \\left( 2 \\cdot ( \\beta_c - \\beta_p ) \\cdot \\tan \\varphi \\right) }

            \\text{Step 2}

            A = \\frac{1 + \\frac{\\gamma \\cdot h_{crit}^{\\prime}}{2 \\cdot p_o}}{1 + \\frac{\\gamma \\cdot h_{crit}}{2 \\cdot p_o}}

            \\text{if } A \\cdot q_{p,(1)} > q_c \\implies A \\cdot q_{p,(1)} = q_c

            \\text{Step 3}

            q_{p,j+1} = q_{p,j} + \\left[ A \\cdot q_{p,(1),j+1} - q_{p,j} \\right] \\cdot \\frac{d}{D}

            \\text{if } q_{p,j+1} > q_{p,(1),j+1} \\implies q_{p,j+1} = q_{p,(1),j+1}

            \\text{Step 4}

            q_{p,q+1} = q_{p,q} + \\left[ (q_{p,j+1})_{q+1} - q_{p,q} \\right] \\cdot \\frac{d}{D}

            \\text{According to Van Impe:}

            q_{p,q+1} = q_{p,q} + 2 \\cdot \\left[ (q_{p,j+1})_{q+1} - q_{p,q} \\right] \\cdot \\frac{d}{D}

            \\text{if } q_{p,q+1} > (q_{p,j+1})_{q+1} \\implies q_{p,q+1} = (q_{p,j+1})_{q+1}


        For numerical stability, rows with zero cone resistance at the top of the cone resistance trace are discarded.

        :param pile_diameter: Diameter of the pile as a multiple of 0.2m
        :param vanimpecorrection: Boolean determining whether the upward correction according to De Beer's original paper (default) or Van Impe (multiplier of 2) needs to be taken into account.
        :param hcrit: :math:`h_{crit}` adopted for De Beer's calculation (based on the mechanical cone). Default=0.2m

        :return: Returns a dataframe `calc` with the different correction stages
        """
        calc = deepcopy(self.calculation_data)
        calc = calc[calc['qc [MPa]'] > 0].reset_index(drop=True)

        # ----------------------------------------------------
        # Step 1: Shallow depth failure surface correction
        # ----------------------------------------------------

        # Calculate phi according to Equation 23
        calc['phi [deg]'] = np.rad2deg(
            self.phi_func()(1000 * calc['qc [MPa]'] / calc['Effective vertical stress [kPa]']))
        # Determine the values of the normalised depths h/d and h/D
        calc['h/d [-]'] = calc['z [m]'] / self.diameter_cone
        calc['h/D [-]'] = calc['z [m]'] / pile_diameter
        # Find values of beta for cone penetration test and pile according to Equation 60
        for i, row in calc.iterrows():
            try:
                root = brentq(
                    f=self.optimisation_func,
                    a=0,
                    b=0.5 * np.pi,
                    args=(row['h/d [-]'], np.deg2rad(row['phi [deg]'])))
            except:
                root = 0.5 * np.pi
            calc.loc[i, 'beta_c [rad]'] = root
            try:
                root = brentq(
                    f=self.optimisation_func,
                    a=0,
                    b=0.5 * np.pi,
                    args=(row['h/D [-]'], np.deg2rad(row['phi [deg]'])))
            except:
                root = 0.5 * np.pi
            calc.loc[i, 'beta_p [rad]'] = root
        # Apply Equation 62 to obtain qp
        calc['qp [MPa]'] = calc['qc [MPa]'] / \
            (np.exp(
              2 *
              (calc['beta_c [rad]'] - calc['beta_p [rad]']) *
              np.tan(np.deg2rad(calc['phi [deg]']))))

        # ----------------------------------------------------
        # Step 2: Stress level correction
        # ----------------------------------------------------

        calc['A qp [MPa]'] = list(map(lambda _qp, _po, _gamma, _qc: min(_qc, self.stress_correction(
            qc=_qp, po=_po, diameter_pile=pile_diameter, diameter_cone=self.diameter_cone, gamma=_gamma, hcrit=hcrit)),
                           calc['qp [MPa]'], calc['Effective vertical stress [kPa]'], calc['Effective unit weight [kN/m3]'], calc['qc [MPa]']))

        # --------------------------------------------------------------------
        # Step 3: Corrections for transition from weaker to stronger layers
        # --------------------------------------------------------------------
        for i, row in calc.iterrows():
            if i > 0:
                calc.loc[i, 'qp,j+1 [MPa]'] = min(row['A qp [MPa]'] ,
                    calc.loc[i-1, 'qp,j+1 [MPa]'] + \
                    (row['A qp [MPa]'] - calc.loc[i-1, 'qp,j+1 [MPa]']) * \
                    (self.diameter_cone / pile_diameter))
            else:
                calc.loc[i, 'qp,j+1 [MPa]'] = 0

        # --------------------------------------------------------------------
        # Step 4: Corrections for transition from stronger to weaker layers
        # --------------------------------------------------------------------
        if vanimpecorrection:
            coefficient = 2.0
        else:
            coefficient = 1.0

        qu = np.zeros(len(calc['z [m]']))
        # Assign the last value of qd as the starting value of qu
        qu[-1] = calc['qp,j+1 [MPa]'].iloc[-1]
        for i, _qd in enumerate(calc['qp,j+1 [MPa]']):
            if i != 0:
                qu[-1 - i] = min(
                    qu[-i] +
                    coefficient * (calc['qp,j+1 [MPa]'].iloc[-1 - i] - qu[-i]) *
                    (self.diameter_cone / pile_diameter),
                    calc['qp,j+1 [MPa]'].iloc[-1 - i]
                )
        calc['qp,q+1 [MPa]'] = qu

        # --------------------------------------------------------------------
        # Step 5: Averaging to 1D below the reference level
        # --------------------------------------------------------------------
        for i, row in calc.iterrows():
            try:
                _window_data = calc[
                    (calc['z [m]'] >= row['z [m]']) &
                    (calc['z [m]'] <= (row['z [m]'] + pile_diameter))]
                calc.loc[i, "qb [MPa]"] = min(
                    row['qp,q+1 [MPa]'],
                    _window_data['qp,q+1 [MPa]'].mean())
            except:
                calc.loc[i, "qb [MPa]"] = row['qp,q+1 [MPa]']

        return calc

    def plot_base_resistance(self, selected_depth=None, show_standard_diameters=False,
                             plot_title=None, plot_height=800, plot_width=600,
                             plot_margin=dict(t=100, l=50, b=50), show_fig=True,
                             x_range=None, x_tick=None, y_range=None, y_tick=None,
                             legend_orientation='h', legend_x=0.05, legend_y=-0.05):
        """
        Plots the base resistance construction according to De Beer

        :param selected_depth: Depth at which the base resistance is requested (default=None for no such depth)
        :param show_standard_diameters: Boolean determining whether the traces of De Beer base unit base resistance computed for diameters which are multiples of 0.2m need to be shown or not (default=False)
        :param plot_title: Title of the plot (default=None)
        :param plot_height: Height for the plot (default=800px)
        :param plot_width: Width of the plot (default=600px)
        :param plot_margin: Margin for the plot. (default=``dict(t=100, l=50, b=50)``
        :param show_fig: Boolean determining whether this figure is shown or not (default=``True``)
        :param x_range: Plotting range for the cone resistance (default=None for Plotly defaults)
        :param x_tick: Tick interval for the cone resistance (default=None for Plotly defaults)
        :param y_range: Plotting range for the depth (default=None for Plotly defaults)
        :param y_tick: Tick interval for the depth (default=None for Plotly defaults)
        :param legend_orientation: Orientation of the legend (default=``'h'`` for horizontal)
        :param legend_x: x Position of the legend (default=0.05 for 5% from plot left edge)
        :param legend_y: y Position of the legend (default=-0.05 for 5% below plot bottom)
        :return: Creates the Plotly figure ``base_plot`` as an attribute of the object
        """
        self.base_plot = subplots.make_subplots(rows=1, cols=1, print_grid=False)
        traceqc = go.Scatter(x=self.qc_raw, y=self.depth_raw,
                             showlegend=True, mode='lines', name=r'$ q_c $')
        self.base_plot.append_trace(traceqc, 1, 1)
        traceqb = go.Scatter(x=self.qb, y=self.depth,
                             showlegend=True, mode='lines', name=r'$ q_{b} $')
        self.base_plot.append_trace(traceqb, 1, 1)
        if show_standard_diameters:
            trace_diameter_1 = go.Scatter(x=self.calc_1['qb [MPa]'], y=self.depth,
                                 showlegend=True, mode='lines', name=r'$ q_{b,%.1fm} $' % self.diameter_1)
            self.base_plot.append_trace(trace_diameter_1, 1, 1)
            trace_diameter_2 = go.Scatter(x=self.calc_2['qb [MPa]'], y=self.depth,
                                          showlegend=True, mode='lines', name=r'$ q_{b,%.1fm} $' % self.diameter_2)
            self.base_plot.append_trace(trace_diameter_2, 1, 1)
        if selected_depth is not None:
            qb_selected = np.interp(selected_depth, self.depth_qb, self.qb)
            traceqb_selected = go.Scatter(
                x=[qb_selected, ], y=[selected_depth, ], showlegend=False, mode='markers',
                marker=dict(size=10, color='red', line=dict(width=2, color='black')))
            self.base_plot.append_trace(traceqb_selected, 1, 1)

        self.base_plot['layout']['xaxis1'].update(
            title=r'$ q_c, \ q_b \ \text{[MPa]} $', side='top', anchor='y',
            range=x_range, dtick=x_tick)
        if y_range is None:
            self.base_plot['layout']['yaxis1'].update(title=r'$ z \ \text{[m]} $', autorange='reversed', dtick=y_tick)
        else:
            self.base_plot['layout']['yaxis1'].update(
                title=r'$ z \ \text{[m]} $', range=y_range, dtick=y_tick)
        self.base_plot['layout'].update(height=plot_height, width=plot_width,
                                        title=plot_title,
                                        margin=plot_margin,
                                        legend=dict(orientation=legend_orientation, x=legend_x, y=legend_y))
        if show_fig:
            iplot(self.base_plot, filename='baseplot', config={'showLink': False})


    def correct_shaft_qc(self, cone_type="E"):
        """
        Corrects the cone resistance for the effect of the cone type according to Belgian practice.
        A correction factor is applied for the mechanical cones M1, M2 and M4

        .. figure:: images/conetype_correction_be.png
            :figwidth: 500.0
            :width: 450.0
            :align: center

            Correction factors to be used according to Belgian practice

        :param cone_type: Cone type. Select from 'M1', 'M2', 'M4', 'E', 'U'
        """
        if cone_type not in ['M1', 'M2', 'M4', 'E', 'U']:
            raise ValueError("Cone type not recognised. Select from 'M1', 'M2', 'M4', 'E', 'U'")

        for i, row in self.layering.iterrows():

            if (cone_type == "M1" or cone_type == "M2") and row['Tertiary clay']:
                self.qc_corrected = self.qc_raw / 1.3
            elif (cone_type == "M4") and row['Tertiary clay']:
                self.qc_corrected = self.qc_raw / 1.15
            else:
                self.qc_corrected = self.qc_raw

    def calculate_average_qc(self, qc_avg_override=None):
        """
        Calculates the average cone resistance in each layer for shaft resistance calculation.
        If `qc_avg_override` is used, the calculation is discarded and the specified value is set

        :param qc_avg_override: List with average qc values in MPa in the layers with soil types specified in `set_soil_type`
        """
        if qc_avg_override is not None:
            self.layering["qc avg [MPa]"] = qc_avg_override
        else:
            for i, row in self.layering.iterrows():
                self.layering.loc[i, "qc avg [MPa]"] = np.nan_to_num(self.qc_corrected[
                    (self.depth_raw >= row["Depth from [m]"]) &
                    (self.depth_raw <= row["Depth to [m]"])]).mean()

    def calculate_unit_shaft_friction(self):
        """
        Calculates the unit shaft friction according to the Belgian practice.

        Note the importance of using correct units. The cone resistance is provided in MPa whereas the unit shaft
        friction is calculated in kPa

        .. math::
            q_s = 1000 \\cdot \\eta_p^* \\cdot q_{c,avg}

        .. figure:: images/unit_shaft_friction_be.png
            :figwidth: 500.0
            :width: 450.0
            :align: center

            Unit shaft friction according to Belgian practice
        """

        for i, row in self.layering.iterrows():
            if row['Soil type'] == 'Clay':
                if row['qc avg [MPa]'] <= 4.5:
                    self.layering.loc[i, "qs [kPa]"] = 1000 * (1 / 30) * row['qc avg [MPa]']
                else:
                    self.layering.loc[i, "qs [kPa]"] = 150
            elif row['Soil type'] == 'Loam (silt)':
                if row['qc avg [MPa]'] <= 6:
                    self.layering.loc[i, "qs [kPa]"] = 1000 * (1 / 60) * row['qc avg [MPa]']
                else:
                    self.layering.loc[i, "qs [kPa]"] = 100
            elif (row['Soil type'] == 'Sandy clay / loam (silt)') or (row['Soil type'] == 'Clayey sand / loam (silt)'):
                if row['qc avg [MPa]'] <= 10:
                    self.layering.loc[i, "qs [kPa]"] = 1000 * (1 / 80) * row['qc avg [MPa]']
                else:
                    self.layering.loc[i, "qs [kPa]"] = 125
            elif row['Soil type'] == 'Sand':
                if row['qc avg [MPa]'] <= 10:
                    self.layering.loc[i, "qs [kPa]"] = 1000 * (1 / 90) * row['qc avg [MPa]']
                elif 10 < row['qc avg [MPa]'] <= 20:
                    self.layering.loc[i, "qs [kPa]"] = 110 + 4 * (row['qc avg [MPa]'] - 10)
                else:
                    self.layering.loc[i, "qs [kPa]"] = 150
            else:
                warnings.warn("Unrecognized soil type (%s) for layer %i" % (row['Soil type'], i+1))
                self.layering.loc[i, "qs [kPa]"] = np.nan

    def plot_unit_shaft_friction(self, plot_title=None, plot_height=800, plot_width=600,
                                 plot_margin=dict(t=100, l=50, b=50), show_fig=True,
                                 x_ranges=(None, (0, 160)), x_ticks=(None, None), y_range=None, y_tick=None,
                                 legend_orientation='h', legend_x=0.05, legend_y=-0.05):
        """
        Plots the qc averaging and the unit shaft friction following from this

        :param plot_title: Title of the plot (default=None)
        :param plot_height: Height for the plot (default=800px)
        :param plot_width: Width of the plot (default=600px)
        :param plot_margin: Margin for the plot. (default=``dict(t=100, l=50, b=50)``
        :param show_fig: Boolean determining whether this figure is shown or not (default=``True``)
        :param x_ranges: Plotting range for the cone resistance (default=None for Plotly defaults)
        :param x_ticks: Tick interval for the cone resistance (default=None for Plotly defaults)
        :param y_range: Plotting range for the depth (default=None for Plotly defaults)
        :param y_tick: Tick interval for the depth (default=None for Plotly defaults)
        :param legend_orientation: Orientation of the legend (default=``'h'`` for horizontal)
        :param legend_x: x Position of the legend (default=0.05 for 5% from plot left edge)
        :param legend_y: y Position of the legend (default=-0.05 for 5% below plot bottom)
        :return: Creates the Plotly figure ``unit_shaft_plot`` as an attribute of the object
        """
        self.unit_shaft_plot = subplots.make_subplots(rows=1, cols=2, print_grid=False, shared_yaxes=True)
        trace_qc = go.Scatter(x=self.qc_raw, y=self.depth_raw,
                             showlegend=True, mode='lines', name=r'$ q_c $')
        self.unit_shaft_plot.append_trace(trace_qc, 1, 1)
        trace_qc_avg = go.Scatter(
            x=np.insert(np.array(self.layering['qc avg [MPa]']),
                        np.arange(len(self.layering['qc avg [MPa]'])),
                        self.layering['qc avg [MPa]']),
            y=np.insert(np.array(self.layering['Depth to [m]']),
                        np.arange(len(self.layering['Depth from [m]'])),
                        self.layering['Depth from [m]']),
            showlegend=True, mode='lines', name=r'$ q_{c,avg} $')
        self.unit_shaft_plot.append_trace(trace_qc_avg, 1, 1)
        trace_qs = go.Scatter(
            x=np.insert(np.array(self.layering['qs [kPa]']),
                        np.arange(len(self.layering['qs [kPa]'])),
                        self.layering['qs [kPa]']),
            y=np.insert(np.array(self.layering['Depth to [m]']),
                        np.arange(len(self.layering['Depth from [m]'])),
                        self.layering['Depth from [m]']),
            showlegend=True, mode='lines', name=r'$ q_{s} $')
        self.unit_shaft_plot.append_trace(trace_qs, 1, 2)
        self.unit_shaft_plot['layout']['xaxis1'].update(
            title=r'$ q_c \ \text{[MPa]} $', side='top', anchor='y',
            range=x_ranges[0], dtick=x_ticks[0])
        self.unit_shaft_plot['layout']['xaxis2'].update(
            title=r'$ q_s \ \text{[kPa]} $', side='top', anchor='y',
            range=x_ranges[1], dtick=x_ticks[1])
        if y_range is None:
            self.unit_shaft_plot['layout']['yaxis1'].update(title=r'$ z \ \text{[m]} $', autorange='reversed', dtick=y_tick)
        else:
            self.unit_shaft_plot['layout']['yaxis1'].update(
                title=r'$ z \ \text{[m]} $', range=y_range, dtick=y_tick)
        self.unit_shaft_plot['layout'].update(height=plot_height, width=plot_width,
                                        title=plot_title,
                                        margin=plot_margin,
                                        legend=dict(orientation=legend_orientation, x=legend_x, y=legend_y))
        if show_fig:
            iplot(self.unit_shaft_plot, filename='shaftplot', config={'showLink': False})


    def set_shaft_base_factors(self, alpha_b_tertiary_clay, alpha_b_other,
                               alpha_s_tertiary_clay, alpha_s_other):
        """
        Sets the shaft and base factors according to Belgian practice.
        The factors are not determined automatically but need to be specified by the user based on the
        Belgian practice. The factors are obtained by fitting the results from static pile load tests.
        Factors for tertiary OC clay and other soil types need to be defined.

        :param alpha_b_tertiary_clay: Base factor for overconsolidated tertiary clay
        :param alpha_b_other: Base factor for other soil types
        :param alpha_s_tertiary_clay: Shaft factor for overconsolidated tertiary clay
        :param alpha_s_other: Shaft factor for other soil types

        """
        for i, row in self.layering.iterrows():
            if row['Tertiary clay']:
                self.layering.loc[i, "alpha_s"] = alpha_s_tertiary_clay
                self.layering.loc[i, "alpha_b"] = alpha_b_tertiary_clay
            else:
                self.layering.loc[i, "alpha_s"] = alpha_s_other
                self.layering.loc[i, "alpha_b"] = alpha_b_other

    def calculate_pile_resistance(self, pile_penetration, base_area, circumference, beta_base=1, lambda_base=1):
        """
        Calculates the pile capacity for a given penetration. In Eurocode 7 terms, this is the calculated pile resistance.

        The unit base resistance according to De Beer is interpolated from the previous calculation and correction
        factors for material, base shape and enlarged bases are taken into account. Note that the pile base area entered
        by the user determines whether a tubular pile behaves in a coring or plugged manner.
        The shaft resistance is calculated based on the unit shaft resistance calculated previously.


        :param pile_penetration: Pile penetration below the soil surface [m]
        :param base_area: Area of the pile base to be used for the base resistance calculation [m2]
        :param circumference: Circumference of the pile shaft [m]
        :param beta_base: Beta factor taking into account the shape of the pile base (default=1 for circular debeer)
        :param lambda_base: Lambda factor taking into account enlarged pile bases (default=1 for uniform cross-sections)


        .. math::
            R_b = \\alpha_b \\cdot \\epsilon_b \\cdot \\beta \\cdot \\lambda \\cdot A_b \\cdot q_b

            \\epsilon_b =
                  \\begin{cases}
                    \\max \\left( 1 - 0.01 \\cdot \\left( \\frac{D_{b,eq}}{D_{CPT}} - 1 \\right); 0.476 \\right)       & \\quad \\text{in tertiary OC clay}\\\\
                    1  & \\quad \\text{in all other soil types}
                  \\end{cases}

            R_s = \\kappa_s \\cdot \\Sum \\left( \\alpha_{s,i} \\cdot h_i \\cdot q_{s,i} \\right)

            R_c = R_s + R_b

        :returns: Sets the following attributes of the DeBeerCalculation object:

            - 'qb': Unit base resistance (:math:`q_b`)  [:math:`kPa`]
            - 'epsilon_b': Factor epsilon_b used in the base resistance calculation
            - 'Rb': Base resistance (:math:`R_b`)  [:math:`kN`]
            - 'Rs': Shaft resistance (:math:`R_s`)  [:math:`kN`]
            - 'Rc': Calculated pile resistance (:math:`R_c`)  [:math:`kN`]
            - 'capacity_calc': Pandas dataframe with the data used for the calculation, contains the components of shaft resistance

        """
        # Validation
        if pile_penetration > self.depth.max():
            raise ValueError("Pile penetration of %.2fm is greater than maximum CPT depth of %.2fm" % (
                pile_penetration, self.depth.max()
            ))
        # Prepare calculation data
        self.capacity_calc = deepcopy(self.layering[self.layering["Depth from [m]"] < pile_penetration])
        self.capacity_calc["Depth to [m]"].iloc[-1] = pile_penetration
        self.capacity_calc['Layer thickness [m]'] = self.capacity_calc['Depth to [m]'] - self.capacity_calc['Depth from [m]']
        # Base resistance calculation
        self.qb_selected = np.interp(pile_penetration, self.depth_qb, self.qb)
        if self.capacity_calc['Tertiary clay'].iloc[-1]:
            self.epsilon_b = max(1 - 0.01 * ((self.diameter_pile / self.diameter_cone) - 1), 0.476)
        else:
            self.epsilon_b = 1
        self.Rb = self.capacity_calc['alpha_b'].iloc[-1] * self.epsilon_b * beta_base * \
                          lambda_base * base_area * (1000 * self.qb_selected)

        # Shaft resistance calculation
        self.capacity_calc["Rs,i [kN]"] = circumference * self.capacity_calc['alpha_s'] * \
                                      self.capacity_calc['Layer thickness [m]'] * self.capacity_calc['qs [kPa]']
        self.Rs = self.capacity_calc["Rs,i [kN]"].sum()
        self.Rc = self.Rs + self.Rb
